using System;
using System.IO;
using System.Xml;
using System.Net;
using System.Collections;
using System.Collections.Generic;

namespace XUIML.Data
{
	/// <summary> The root node of the dataset </summary>
	public class Document
	{
		#region Public Properties
		
		/// <summary> The complete file path of the XUIML Document. </summary>
		public string FilePath { get; private set; }

		/// <summary> The node where the nodes . </summary>
		public Node ParentNode { get; private set; }

		/// <summary>  List of child nodes. </summary>
		public List<Node> Nodes { get; private set; }
		
		/// <summary> The parent document. </summary>
		public Document Parent { get; private set; }

		/// <summary>  List of child documents. </summary>
		public List<Document> Children {get; private set; }
		
		/// <summary> Indicates whether the document is updated or not. </summary>
		public bool Updated { get; protected set; }

		/// <summary> Indicates the frequency of updates. </summary>
		public float UpdateFrequency { get; protected set; }

		#endregion
		
		#region Public Constructor
		
		/// <summary> Initializes a new instance of the <see cref="XUIML.DataSet+Document"/> class. </summary>
		/// <param name="filePath">Path.</param>
		public Document (string filePath, Node node = null, Document parent = null)
		{
			// Store the given values
			this.FilePath = filePath; this.Parent = parent; this.ParentNode = node;
			
			// Initilize the lists of child documents and child nodes
			this.Children = new List<Document>(); this.Nodes = new List<Node>();
			
			// Make sure the filepath is not null
			if (filePath == null || filePath.Length == 0) throw new ArgumentNullException("filePath");
			
			// If the parent is not null, add this instance to its child list and update the data NodeSet variable
			if (this.Parent != null) { this.Parent.Children.Add(this); }
			
			// Save the document reference
			if (this.ParentNode != null) { this.ParentNode.Document = this;}
		}
		
		#endregion
		
		#region Public Methods
		
		/// <summary> Deserializes the XUIML Document from a given file. </summary>
		/// <param name="filePath"> The file path of the XUIML document. </param>
		public void Deserialize()
		{
			// Create a XML document
			System.Xml.XmlDocument document = new System.Xml.XmlDocument();
			
			// If it is a local resource, remove the extension
			string localPath = Path.GetDirectoryName(FilePath) + "/" + Path.GetFileNameWithoutExtension(FilePath);

			// TODO Make the load of XML resources engine-independent
			document.Load(new StringReader(XUIML.Unity.Manager.LoadString(localPath))); 

			// Check that it is a valid XUIML file
			if (document.DocumentElement.Name != "xuiml") throw new Exception("Invalid XUIML file: " + localPath);

			// Deserialize the document main node
			foreach (XmlNode xmlChild in document.DocumentElement.ChildNodes)
			{
				if (ParentNode.Contains(xmlChild.Name))
				{
					// Get the child node
					Node child = (Node)ParentNode[xmlChild.Name];
					
					// If the node is a element of a NodeSet
					if (child.GetType().IsGenericType)
					{
						// Create a variable to store the ID of the child
						string childId = null;
						
						// Try to get the ID from the attribute
						if (xmlChild.Attributes["id"] != null) childId = xmlChild.Attributes["id"].InnerText;
						
						// If no appropiate name has been found, create one
						if (childId == null || childId.Length == 0 || child.Contains(childId))
							childId = xmlChild.Name + child.ChildCount;
						
						// Create a new child node and deserialize it
						Node childNode = child.Add(childId); Serialization.Deserialize(childNode, xmlChild);
						Nodes.Add(childNode);
					}
					else throw new Exception("Invalid node name '" + xmlChild.Name + "' for " + ParentNode.Name);
				}
			}
		}

		/// <summary> Serializes the XUIML Document. </summary>
		/// <param name="timesUpdated"> The minimum number of times the node must have been updated. </param>
		public void Serialize (float timesUpdated = 1)
		{
			// Create a XML document
			System.Xml.XmlDocument document = new System.Xml.XmlDocument();

			// Serialize the node
			document.AppendChild(Serialization.Serialize(this.ParentNode, document, timesUpdated));

			// Check if the file should be stored locally or uploaded online (HTTP POST)
			if (!FilePath.StartsWith("http")) document.Save(this.FilePath);
			else {
				// use a web client ot upload the data easily
				using(WebClient client = new WebClient())
				{
					// Make sure the encoding is performed in UTF-8
					client.Encoding = System.Text.Encoding.UTF8;

					// Upload the string in a synctronous fashion and 
					string response = client.UploadString (FilePath, document.InnerXml);

					if (!response.StartsWith("OK"))	throw new Exception("Serialization Failed: " + response);
				}
			}
		}

		
		#endregion
	}
}
