using System;
using System.Collections;
using System.Collections.Generic;

namespace XUIML.Data
{
	/// <summary> An abstract class that defines a gnereric XUIML Node.</summary>
	public class Node : IEnumerable
	{
		#region Public events
		
		/// <summary> Defines the structure of event handlers </summary>
		public delegate void EventHandler (Node node);
		
		/// <summary> Occurs when an node is created. </summary>
		public static event EventHandler Created;
		
		/// <summary> Occurs when an update operation is requested. </summary>
		public event EventHandler UpdateRequested;
		
		/// <summary> Occurs when an update operation is performed. </summary>
		public event EventHandler UpdatePerformed;
		
		#endregion

		#region Protected Fields
		
		/// <summary> The dictionary of children nodes. </summary>
		protected IDictionary Children;

		#endregion
		
		#region Public Properties
		
		/// <summary> The name of the node. </summary>
		public string Name { get; protected set; }

		/// <summary> A string locator that univocally identifies the node in the hierarchy. </summary>
		public string Locator { get { if (Parent == null) return Name; else return Parent.Locator + "." + Name; } }

		/// <summary> Gets a child node using a locator. </summary>
		/// <param name="locator"> The name of the child. </param>
		public Node this [string locator]
		{
			get {
				int separatorIndex = locator.IndexOf('.'); if(separatorIndex < 0) return (Node) Children[locator];
				return ((Node) Children[locator.Substring(0, separatorIndex)])[locator.Substring(separatorIndex + 1)];
			}
		}

		/// <summary> Gets the child count. </summary>
		public int ChildCount { get { return Children.Count; } }
		
		/// <summary> Indicates whether the node is updated or not. </summary>
		public bool Updated { get; protected set; }
		
		/// <summary> The parent node instance. </summary>
		public Node Parent { get; protected set; }

		/// <summary> The Document associated with this node. </summary>
		public Document Document { get; set; }

		/// <summary> The Representation of the node. </summary>
		public Representation Representation { get; set; }

		/// <summary> The number of times the node has been updated. </summary>
		public int TimesUpdated { get; protected set; }

		#endregion
		
		#region Public Constructor
		
		/// <summary> Initializes a new instance of the <see cref="XUIML.Dataset+Node"/> class. </summary>
		/// <param name="name"> The name of the node. </param>
		/// <param name="parent"> The parent node. </param>
		public Node (string name, Node parent = null, Document document = null)
		{
			// Store the given values
			this.Name = name; this.Parent = parent; this.Document = document;
			
			// Initialize the internal collections
			this.Children = new Dictionary<string, Node>();

			// Mark the node for its first update
			Updated = false; TimesUpdated = 0;

			// If the NodeSet does not contain a name do not add it to the parent children
			if (this.Parent != null) this.Parent.Add(this);

			// Raise the UpdatePerformed event
			if (Created != null) Created(this);
		}
		
		#endregion

		#region Public Methods
		
		/// <summary> Adds the specified node instance. </summary>
		/// <param name="node"> The node instance to add to the collection. </param>
		public virtual Node Add(Node node)
		{
			// Check the given value
			if (node == null) throw new ArgumentNullException("node");
			if (Children.Contains(node.Name)) throw new Exception ("Duplicated child node name: " + node.Name);
			
			// Add the node to the child list
			Children.Add(node.Name, node);
			
			// Update the parent node value and the update status
			node.Parent = this;

			// Request an update of the entire hierarchy
			RequestUpdate();

			// Retun the node reference (really useful to simplify code in constructors);
			return node;
		}
		
		/// <summary> Adds a new node to th. </summary>
		/// <param name="nodeType"> . </param>
		public virtual Node Add(string nodeName) { return Add(new Node(nodeName)); }
		
		/// <summary> Indicates whether a node is a child of this node. </summary>
		/// <param name="nodeName"> The name of the node to check. </param>
		/// <returns> true if the node has a child with the given name, false otherwise. </returns>
		public bool Contains(string nodeName) { return Children.Contains(nodeName); }

		/// <summary> Looks for an specific ancestor node. </summary>
		/// <param name="name"> (Optional) The name of the ancestor node. </param>
		/// <typeparam name="NodeType">The node type to look for. </typeparam>
		public NodeType Ancestor<NodeType>(string name = null) where NodeType : Node
		{
			// Start using the parent node
			Node node = this.Parent;

			// Search for the appropriate node
			while (node != null)
			{
				// Check if the node satisfies the given conditions
				if(node is NodeType && (name == null || node.Name == name)) return (NodeType)node;

				// Otherwise, continue with the parent node
				node = node.Parent;
			}

			// If no appropriate node has been found, return a null reference
			return null;
		}

		/// <summary> Looks for an specific descendant node. </summary>
		/// <param name="name"> (Optional) The name of the descendant node. </param>
		/// <typeparam name="NodeType">The node type to look for. </typeparam>
		public NodeType Descendant<NodeType>(string name = null) where NodeType : Node
		{
			// Search for the appropriate node
			foreach (Node node in Children)
			{
				// Check if the node satisfies the given conditions
				if(node is NodeType && (name == null || node.Name == name)) return (NodeType)node;

				// Recursivelly look for the appropriate node
				NodeType result = node.Descendant<NodeType>(name); if (result != null) return result;
			}

			// If no appropriate node has been found, return a null reference
			return null;
		}

		/// <summary> Returns an enumerator that iterates through a collection. </summary>
		/// <returns>The enumerator.</returns>
		public IEnumerator GetEnumerator () { return Children.Values.GetEnumerator(); }
		
		/// <summary> Updates the node. </summary>
		/// <param name="forced"> Indicates whether the update must be forced on not. </param>
		public virtual void Update(bool forced = false)
		{
			// Check if the update operation is really required
			if (this.Updated && !forced) return;

			// Update the node representation
			if (this.Representation != null) this.Representation.Update();

			// Mark this node as updated
			this.Updated = true; TimesUpdated++;

			// Update the children nodes
			foreach (Node child in Children.Values) child.Update();

			// Raise the UpdatePerformed event
			if (UpdatePerformed != null) UpdatePerformed(this);
		}

		/// <summary> Request an update of the parent and related nodes. </summary>
		public virtual void RequestUpdate(Node origin = null)
		{
			// Set the update variable of this node to false
			this.Updated = false;

			// Raise the UpdateRequested event
			if (UpdateRequested != null) UpdateRequested(this);

			// Recursivelly sets the updated value throughout the entire hierarchy.
			if (this.Parent != null && this.Parent.Updated) this.Parent.RequestUpdate();
		}

		/// <summary> Removes the specified node instance. </summary>
		/// <param name="node"> The node instance to remove the collection. </param>
		public virtual void Remove(Node node)
		{
			if (!Children.Contains(node.Name)) throw new Exception ("Invalid node name: " + node.Name);

			// Remove the node from the child list
			Children.Remove(node.Name);

			// Update the parent node value
			node.Parent = null;

			// Destroy the node
			node.Destroy();
		}


		public virtual void Destroy()
		{
			if (this.Parent != null) this.Parent.Remove(this);
			if (Representation != null) { Representation.Destroy(); Representation = null; }
		}

		/// <summary> Obtains a see cref="XUIML.Data.Node"/> instance from a string. </summary>
		/// <param name="value"> The string with the see cref="XUIML.Data.Node"/> instance data.</param>
		public virtual void FromString(string value) {}

		/// <summary> Returns a string that represents the current <see cref="XUIML.Data.Node"/> instance. </summary>
		/// <returns> A string that represents the current <see cref="XUIML.Data.Node"/> instance. </returns>
		public override string ToString () { return null; }

		#endregion
	}
}
