using System;
using System.Collections;
using System.Collections.Generic;

namespace XUIML.Data
{
	/// <summary> Represents a complete XUIML dataset. </summary>
	public sealed class NodeSet<NodeType> : Node, IEnumerable<NodeType> where NodeType : Node
	{
		#region Public Properties
		
		/// <summary> Gets the child Node with the specified name. </summary>
		/// <param name="name"> The name of the child node. </param>
		public new NodeType this [string name] { get { return (NodeType)Children[name]; } }

		/// <summary> Gets the child Node with the specified name. </summary>
		/// <param name="index> The indec of the child node. </param>
		public NodeType this [int index]
		{
			get {
				IEnumerator<NodeType> en = GetEnumerator(); 
				for (int i = 0; i<= index; i++) en.MoveNext();
				return en.Current;
			}
		}

		#endregion
		
		#region Public Properties
		
		/// <summary> Initializes a new instance of the <see cref="XUIML.DataSet+Node+NodeSet`1"/> class. </summary>
		/// <param name="name"> The name of the NodeSet. </param>
		/// <param name="parent"> The parent node. </param>
		public NodeSet(string name, Node parent = null) : base (name, parent)
		{ Children = new Dictionary<string, NodeType>(); }
		
		#endregion
		
		#region Public Methods
		
		/// <summary> Adds the specified node instance. </summary>
		/// <param name="node"> The node instance to add to the collection. </param>
		public override Node Add(Node node)
		{
			if (node == null) throw new ArgumentNullException("node");

			Children.Add(node.Name, (NodeType)node);
		
			RequestUpdate();

			return node;
		}
		
		/// <summary> Adds the specified node instance. </summary>
		/// <param name="node"> The node instance to add to the collection. </param>
		public override Node Add(string nodeName)
		{
			if (nodeName == null) nodeName = this.Name + this.Children.Count;
			return (NodeType)Activator.CreateInstance(typeof(NodeType), new object[]{nodeName, this}, null); }

		public int IndexOf (string name)
		{ int i = 0; foreach (string childName in Children.Keys) { if (childName == name) return i; i++;} return -1; }

		public int IndexOf (NodeType node)
		{ int i = 0; foreach (NodeType child in Children.Values) { if (child == node) return i; i++;} return -1; }


		/// <summary> Returns an enumerator that iterates through a collection. </summary>
		/// <returns> The enumerator. </returns>
		public new IEnumerator<NodeType> GetEnumerator()
		{ return ((Dictionary<string, NodeType>)Children).Values.GetEnumerator(); }

		#endregion
	}
}