using System;
using System.IO;
using System.Xml;
using System.Globalization;
using System.Collections;
using System.Collections.Generic;

namespace XUIML.Data
{
	/// <summary> Handles the conversion of the XUIML Model from and to other data representations. </summary>
	public sealed class Serialization
	{
		#region Public Constants

		public const char WhiteSpaceSeparator = ' ';
		public const char TabSeparator = '\t';

		public const char DigitMinValue = '0';
		public const char DigitMaxValue = '9';
		public const char LetterMinValue = 'A';
		public const char LetterMaxValue = 'z';


		#endregion

		#region Public Subclasses

		public sealed class XML
		{
			
		}

		public sealed class JSON
		{
			public const char DefinitionSeparator = ':';
			public const char ValueSeparator = ',';
		}

		public sealed class CSV
		{
			public const char ValueSeparator = ',';

			/// <summary> Recreates a XUIMl node from its CSV representation. </summary>
			/// <param name="node"> The node reference. </param>
			/// <param name="text"> The text with the node data. </param>
			public static void Deserialize (Node node, string text)
			{
				// Obtain the list of child nodes
				List<Node> children = new List<Node>();
				foreach (object childObject in node) children.Add((Node)childObject);

				// TODO Create a version that  allows the deserializztion of objects to a NodeSet 

				// Initialize the required variables to parse the text
				int charIndex = 0;  string buffer= ""; char character; int childIndex = 0; Node child;

				// Begin parsing the text as CSV data
				while (charIndex < text.Length)
				{
					character = text[charIndex++];
					child = children[childIndex];

					if (character == WhiteSpaceSeparator) continue;

					// 
					if (character == ValueSeparator || charIndex == text.Length)
					{
						if (charIndex == text.Length) buffer += character;
						if (child is Text) ((Text)child).Value = buffer;
						else if (child is Number) ((Number)child).Value = float.Parse(buffer);
						else if (child is Boolean) ((Boolean)child).Value = bool.Parse(buffer);

						buffer=""; childIndex++; continue;
					}

					// Add the character to the buffer
					buffer += character;
				}
			}
		}

		#endregion

		#region Public Fields

		public static char[] Separators = {WhiteSpaceSeparator, TabSeparator };

		public static string CurrentLanguage = null;

		#endregion
		
		#region Public Methods

		public static bool IsDigit(char character) { return character >= DigitMinValue && character <= DigitMaxValue; }
		
		public static bool IsLetter(char character) { return char.IsLetter(character); }
		
		public static bool IsLetterOrDigit(char character) { return IsDigit(character) || IsLetter(character); }
		
		public static bool IsSeparator(char character, char[] separators = null)
		{
			if (separators == null) separators = Separators;
			foreach (char separator in separators) if (character == separator) return true;
			return false;
		}

		public static void Deserialize (Node node, string text)
		{
			// TEMPORAL CSV
			CSV.Deserialize(node, text);

//			// TEMPORAL just JSON
//			int charIndex = 0; string buffer= "";
//
//
//			while (true)
//			{
//				char character = text[charIndex];
//
//				if (character > 'A' && character < 'z') buffer += character;
//				else if (character == JSON.DefinitionSeparator) value
//
//				charIndex++;
//			}
		}

		/// <summary> Deserializes a Node from an XML Node. </summary>
		/// <param name="xmlNode"> The XML node. </param>
		public static void Deserialize(Node node, System.Xml.XmlNode xmlNode)
		{
			//DebugConsole.Log("Serialized " + xmlNode.Name);
			if (node is Text) ((Text)node).Value = xmlNode.InnerText;
			else if (node is Boolean) ((Boolean)node).Value = bool.Parse(xmlNode.InnerText);
			else if (node is Index) ((Index)node).Value = int.Parse(xmlNode.InnerText);
			else if (node is Number) ((Number)node).Value = float.Parse(xmlNode.InnerText);

			// Manage the XML attributes
			foreach (System.Xml.XmlAttribute attribute in xmlNode.Attributes)
				if (node.Contains(attribute.Name))
					((Node)node[attribute.Name]).FromString(attribute.InnerText);

			// Manage the XML elements
			foreach (System.Xml.XmlNode xmlChild in xmlNode.ChildNodes)
			{
				// Ignore XML comments
				if (xmlChild.Name == "#comment") continue;
				
				// Manage inclusions
				if (xmlChild.Name == "include")
				{
					// Check if the HREF value exists
					if (xmlChild.Attributes.GetNamedItem("href") == null) 
						throw new Exception("HREF undefined");

					// Check if there if a language conditions
					if (xmlChild.Attributes.GetNamedItem("language") != null)
					{
						string languageText = xmlChild.Attributes.GetNamedItem("language").InnerText;
						string systemLanguage = UnityEngine.Application.systemLanguage.ToString();
						if ((CurrentLanguage == null) && ((languageText == systemLanguage) || (languageText == "default")))
							CurrentLanguage = languageText;
						if (CurrentLanguage != languageText) continue;
					}

					// Search for the first node with an associated document
					Node searchNode = node.Parent;
					while (searchNode.Document == null && searchNode.Parent != null) searchNode = searchNode.Parent;
					Document parentDocument = searchNode.Document;
					
					// Obtain the absolute filepath
					string relativeFilePath = xmlChild.Attributes.GetNamedItem("href").InnerText;
					string filePath = Path.GetDirectoryName(parentDocument.FilePath) + 	"/" + relativeFilePath;
					Document includedDocument = new Document(filePath, node, parentDocument);
					includedDocument.Deserialize();

					// Include nodes are not allowed to include children
					continue;
				}
				
				// Check if the child list contains a node with the same name
				if (node.Contains(xmlChild.Name))
				{
					// Get the child node
					Node child = (Node)node[xmlChild.Name];
					
					// If the node is a element of a NodeSet
					if (child.GetType().IsGenericType)
					{
						// Create a variable to store the ID of the child
						string childId = null;
						
						// Try to get the ID from the attribute
						if (xmlChild.Attributes["id"] != null) childId = xmlChild.Attributes["id"].InnerText;
						
						// If no appropiate name has been found, create one
						if (childId == null || childId.Length == 0 || child.Contains(childId))
							childId = xmlChild.Name + child.ChildCount;
						
						// Create a new child node and deserialize it
						Node childNode = child.Add(childId); Deserialize(childNode, xmlChild);
					}
					else throw new Exception("Invalid node name '" + xmlChild.Name + "' for " + node.Name);
				}
			}
		}


		/// <summary> Serializes the specified node. </summary>
		/// <param name="node"> The node to serialize. </param>
		/// <param name="document"> The XML document that contains the node.</param>
		/// <param name="timesUpdated"> The minimum number of times the node must have been updated. </param>
		public static XmlNode Serialize(Node node, XmlDocument document, float timesUpdated = 1)
		{
			// Avoid Serializing the nodes that have not been updated a given number of times
			if (node.TimesUpdated < timesUpdated) return null; 

			// Create the XML element with the right name
			XmlElement xmlNode; string xmlNodeName = "";
			if (node.Parent.GetType().IsGenericType) xmlNodeName = node.Parent.Name; else xmlNodeName = node.Name;
			xmlNode = document.CreateElement(xmlNodeName);

			// Basic node tipes are converted directly into text nodes 
			if (node is Text || node is Boolean || node is Index || node is Number) xmlNode.InnerText = node.ToString();

			// Check each child node
			foreach (Node child in node)
			{
				// Check if the node is a a Node set
				if (child.GetType().IsGenericType)
				{
					// Check each child of the nodeset
					foreach(Node grandChild in child)
					{ 
						// Once more, check the times the node has been updated
						if (grandChild.TimesUpdated < timesUpdated) continue; 

						// Serialize the subnode
						System.Xml.XmlNode xmlSubNode = Serialize(grandChild, document, timesUpdated);

						// Add the id XML attribute
						XmlAttribute attribute = document.CreateAttribute("id");
						attribute.InnerText = grandChild.Name;
						xmlSubNode.Attributes.Append(attribute);

						// Append the grandchild node
						xmlNode.AppendChild(xmlSubNode);
					}
				}
				else
				{
					// Add the node has an XML attribute
					XmlAttribute attribute = document.CreateAttribute(child.Name);
					attribute.InnerText = child.ToString();
					xmlNode.Attributes.Append(attribute);
				}
			}

			// Return the resulting node
			return xmlNode;
		}
	
		#endregion
	}
}