using XUIML.Data;
using System.Collections.Generic;

namespace XUIML.Data
{
	/// <summary> Describes a behaviour. </summary>
	public sealed class Behavior : Node
	{
		#region Public Subclasses

		/// <summary> Defines the respose before a given event. </summary>
		public sealed class Action : Node
		{
			#region Public Subclasses

			/// <summary> Defines the reaction to a concrete action. </summary>
			public sealed class Reaction : Node
			{
				#region Public Properties

				/// <summary> The name of the node affected by the reaction. </summary>
				public Text Key { get; private set; }

				/// <summary> The operator of the reaction. </summary>
				public Text Operator { get; private set; }

				/// <summary> The value of the reaction. </summary>
				public Text Value { get; private set; }
				
				#endregion
				
				#region Public Constructor
				
				/// <summary> Initializes a new instance of the <see cref="XUIML.Data.Behaviour.Action.Reaction"/>
				/// class. </summary>
				/// <param name="name"> The node name. </param>
				/// <param name="parent"> The parent node. </param>
				public Reaction (string name, Node parent = null, string reaction = null) : base(name, parent)
				{
					this.Key = new Text("key", this);
					this.Operator = new Text("operator", this); 
					this.Value = new Text("value", this);
				}

				#endregion
			}
			
			#endregion

			#region Public Properties

			/// <summary> The reactions associated to the action. </summary>
			public NodeSet<Reaction> Reactions { get; private set; }

			#endregion
			
			#region Public Constructor
			
			/// <summary> Initializes a new instance of the <see cref="XUIMLData.Behaviour.Action"/> class. </summary>
			/// <param name="name"> The node name. </param>
			/// <param name="parent"> The parent node. </param>
			public Action (string name, Node parent = null, string reactions = null) : base(name, parent)
			{ this.Reactions = new NodeSet<Reaction>("reaction", this); }

			#endregion
		}

		#endregion

		#region Private Enumerations
		
		/// <summary> Enumerates he different parsing states. </summary>
		private enum ParseState { Action, ReactionKey, ReactionOperation, ReactionValue};
		
		#endregion

		#region Public Properties

		/// <summary> A text with the actions of the behaviour. </summary>
		public Text ActionsText { get; private set; }

		/// <summary> The different actions that define the behaviour. </summary>
		public NodeSet<Action> Actions { get; private set; }
		
		#endregion
		
		#region Public Constructor
		
		/// <summary> Initializes a new instance of the <see cref="XUIML.Data.Behavior"/> class. </summary>
		/// <param name="name"> The node name. </param>
		/// <param name="parent"> The parent node. </param>
		public Behavior (string name, Node parent = null) : base(name, parent)
		{ this.Actions = new NodeSet<Action>("action", this); }
		
		#endregion

		#region Public Methods
		
		public override void FromString (string value)
		{
			// Make sure the given string can be parsed
			if (value == null && value.Length == 0) return;

			// Extract the different actions
			System.StringSplitOptions noNullStrings = System.StringSplitOptions.RemoveEmptyEntries;
			string[] actionStrings = value.Split(new string[]{"{", "}"}, noNullStrings);

			// Create some local variables
			Action action = null;
			ParseState parseState = ParseState.Action;

			// Check all the string parts
			for (int actionStringIndex = 0; actionStringIndex < actionStrings.Length; actionStringIndex++)
			{
				string actionString = actionStrings[actionStringIndex];

				if (parseState == ParseState.Action)
				{
					// Clean the action name string and create its node
					string[] actionNameStrings = actionString.Split(new string[]{" ", "\t", "\n"}, noNullStrings);
					if (actionNameStrings.Length != 1) throw new System.Exception("Invalid reaction name: [" + actionString + "]");
					action = new Action(actionNameStrings[0], this.Actions);

					// 
					parseState = ParseState.ReactionKey;
				}
				else
				{
					// Extract all the reactions
					string[] reactionStrings = actionString.Split(new string[]{";"}, noNullStrings);

					// Parse each action
					for (int reactionStringIndex = 0;reactionStringIndex < reactionStrings.Length;reactionStringIndex++)
					{
						// Get the reaction string
						string reactionString = reactionStrings [reactionStringIndex];

						// Get the reaction operator
						string reactionOperator;
						if (reactionString.Contains("+=")) reactionOperator = "+=";
						else if (reactionString.Contains("-=")) reactionOperator = "-=";
						else if (reactionString.Contains("=")) reactionOperator = "=";
						else throw new System.Exception("Invalid reaction operator in: [" + reactionString + "]");

						string[] reactionStringParts = reactionString.Split(
							new string[]{reactionOperator, " ", "\t", "\n"}, noNullStrings);
						if (reactionStringParts.Length != 2)
							throw new System.Exception("Invalid reaction: " + reactionString);
					
						Action.Reaction reaction = new Action.Reaction("reaction" + action.Reactions.ChildCount, action.Reactions);
						reaction.Key.Update(reactionStringParts[0]); reaction.Value.Update(reactionStringParts[1]);
						reaction.Operator.Update(reactionOperator);
					}

					// Back in action
					parseState = ParseState.Action;
				}
			}
		}

		#endregion
	}
}


