using UnityEngine;
using System;

namespace XUIML.Data
{
	/// <summary> Describes a Color XUIML Node. </summary>
	public sealed class Color : Node
	{
		#region Public Properties
		
		/// <summary> The Red color component value. </summary>
		public Number R { get; private set; }
		
		/// <summary> The Green color component value. </summary>
		public Number G { get; private set; }

		/// <summary> The Blue color coordinate value. </summary>
		public Number B { get; private set; }

		public Color White
		{ get { Color c = new Color("color"); c.R.Value = 1.0f; c.G.Value = 1.0f; c.B.Value = 1.0f; return c; } }

		#endregion
		
		#region Public Constructors
		
		/// <summary> Initializes a new instance of the <see cref="XUIML.Color"/> class. </summary>
		/// <param name="name"> The node name. </param>
		/// <param name="parent"> The parent node. </param>
		public Color (string name, Node parent = null) : base(name, parent)
		{ this.R = new Number("r", this); this.G = new Number("g", this); this.B = new Number("b", this); }

		/// <summary> Initializes a new instance of the <see cref="XUIML.Data.Color"/> class. </summary>
		/// <param name="name"> The node name. </param>
		/// <param name="parent"> The parent node. </param>
		/// <param name="r">The Red color component. </param>
		/// <param name="g">The Green color component. </param>
		/// <param name="b">The Blue color component. </param>
		public Color (string name, Node parent, Number r, Number g, Number b) : this(name, parent)
		{ this.R = r; this.G = g; this.B = b; }
		
		#endregion

		#region Public Methods

		/// <summary> Updates the node with specific values. </summary>
		///  <param name="color"> the color values. </param>
		public void Update(Color color) { this.R.Value = color.R; this.G.Value = color.G; this.B.Value = color.B; }

		/// <summary> Updates the node with specific values. </summary>
		/// <param name="r"> The Red color component. </param>
		/// <param name="g"> The Green color component. </param>
		/// <param name="b"> The Blue color component. </param>
		public void Update(Number r, Number g, Number b) { this.R.Value = r; this.G.Value = g; this.B.Value = b; }

		/// <summary> Updates the node with specific values. </summary>
		/// <param name="r"> The Red color component. </param>
		/// <param name="g"> The Green color component. </param>
		/// <param name="b"> The Blue color component. </param>
		public void Update(float r, float g, float b) { this.R.Value = r; this.G.Value = g; this.B.Value = b; }

		/// <summary> Obtains the object data from a string representation. </summary>
		/// <param name="value"> The string with the object data.</param>
		public override void FromString (string value) { Serialization.CSV.Deserialize(this, value); }
		
		/// <summary> Returns a string that represents the current object. </summary>
		/// <returns>A string that represents the current object.</returns>
		public override string ToString () { return string.Format ("[{0}, {1}, {2}]", this.R, this.G, this.B); }

		#endregion
	}
}
