namespace XUIML.Data
{
	/// <summary> Describes the contextual information. </summary>
	public sealed class Context : Node
	{
		#region Public Subclasses

		/// <summary> Describes a context condition. </summary>
		public sealed class Condition : Node
		{
			#region Public Properties

			/// <summary> The boolean connector. </summary>
			public Text Connector { get; private set; }

			/// <summary> The name of the node affected by the condition. </summary>
			public Text Key { get; private set; }

			/// <summary> The operator of the condition. </summary>
			public Text Operator { get; private set; }

			/// <summary> The value of the condition. </summary>
			public Text Value { get; private set; }

			/// <summary> The node affected by the condition. </summary>
			public Node Node { get; set; }

			/// <summary> Indicates wheteher the condition is active or not. </summary>
			public bool Active { get; set; }

			#endregion

			#region Public Constructor

			/// <summary> Initializes a new instance of the <see cref="XUIML.Context.Condition"/> class. </summary>
			/// <param name="name"> The node name. </param>
			/// <param name="parent"> The parent node. </param>
			public Condition (string name, Node parent = null) : base(name, parent)
			{
				this.Connector = new Text("connector", this);
				this.Key = new Text("key", this);
				this.Operator = new Text("operator", this);
				this.Value = new Text("value", this);
			}

			#endregion
		}

		#endregion

		#region Private Enumerations
		
		/// <summary> Enumerates he different parsing states. </summary>
		private enum ParseState { Connector, Key, Operation, Value};
		
		#endregion

		#region Public Properties

		/// <summary> The conditions that define the context. </summary>
		public NodeSet<Condition> Conditions { get; private set; }

		/// <summary> Indicates wheteher the context is active or not. </summary>
		public bool Active { get; set; }

		/// <summary> Indicates wheteher the context is integrated or not. </summary>
		public bool Integrated { get; set; }

		#endregion
		
		#region Public Constructor
		
		/// <summary> Initializes a new instance of the <see cref="XUIML.Data.Context"/> class. </summary>
		/// <param name="name"> The node name. </param>
		/// <param name="parent"> The parent node. </param>
		public Context (string name, Node parent = null) : base(name, parent)
		{ this.Conditions = new NodeSet<Condition>("condition", this); Integrated = false; }
		
		#endregion

		#region Public Methods

		/// <summary> Parses the <see cref="XUIML.Data.Context"/> data from a string. </summary>
		/// <param name="value"> The string with the context data. </param>
		public override void FromString (string value)
		{
			// Make sure the given string can be parsed
			if (value == null && value.Length == 0) return;

			// Extract the different condition strings
			System.StringSplitOptions noNullStrings = System.StringSplitOptions.RemoveEmptyEntries;
			string[] conditionStrings = value.Split(new string[]{" ", "\t", ","}, noNullStrings);

			// Create some local variables
			Condition condition;
			ParseState parseState = ParseState.Connector;
			string conditionConnector = "", conditionKey = "", conditionOperator = "", conditionValue = "";

			// Check all the string parts
			for (int conditionStringIndex = 0; conditionStringIndex < conditionStrings.Length; conditionStringIndex++)
			{
				// get the current string part
				string conditionString = conditionStrings[conditionStringIndex];

				// Act differently depending on the current parse state
				switch (parseState)
				{
				case ParseState.Connector:
					// Get the conditional connector
					if (conditionString == "and") conditionConnector = "and";
					else if (conditionString == "or") conditionConnector = "or";
					else { conditionConnector = "and"; conditionStringIndex --; } // Implicit connector 
					parseState = ParseState.Key;
					break;
				case ParseState.Key:

					// Allow the parsing of complete conditions
					if (conditionString.Contains("=") || conditionString.Contains("{") || conditionString.Contains("}"))
					{
						// Get the conditional operator
						if (conditionString.Contains("==")) conditionOperator = "=";
						else if (conditionString.Contains("{=")) conditionOperator = "<=";
						else if (conditionString.Contains("}=")) conditionOperator = ">=";
						else if (conditionString.Contains("!=")) conditionOperator = "!=";
						else if (conditionString.Contains("{")) conditionOperator = "<";
						else if (conditionString.Contains("}")) conditionOperator = ">";
						else if (conditionString.Contains("=")) conditionOperator = "=";
						else throw new System.Exception("Invalid conditional operator in: " + conditionString);

						// Get the condition key and value from the given string
						string[] conditionParts = conditionString.Split(new string[]{"=","{","}","!"}, noNullStrings);
						if (conditionParts.Length == 2) new System.Exception("Invalid condition: " + conditionString);
						conditionKey = conditionParts[0]; conditionValue = conditionParts[1];
						if (conditionValue == "NULL") conditionValue = "";

						// Create the condition node using the stored values
						condition = new Condition("condition" + Conditions.ChildCount, this.Conditions);
						condition.Connector.Update(conditionConnector); condition.Key.Update(conditionKey);
						condition.Operator.Update(conditionOperator); condition.Value.Update(conditionValue);

						//Restart the process
						parseState = ParseState.Connector;
					}
					else { conditionKey = conditionString; parseState = ParseState.Operation; }
					break;
				case ParseState.Operation:
					// Get the conditional operator
					if (conditionString == "==") conditionOperator = "=";
					else if (conditionString == "{=") conditionOperator = "<=";
					else if (conditionString == "}=") conditionOperator = ">=";
					else if (conditionString == "!=") conditionOperator = "!=";
					else if (conditionString == "{") conditionOperator = "<";
					else if (conditionString == "}") conditionOperator = ">";
					else if (conditionString == "=") conditionOperator = "=";
					else throw new System.Exception("Invalid conditional operator in: [" + conditionString + "]");

					parseState = ParseState.Value;
					break;
				case ParseState.Value:
					// Get the conditional value
					conditionValue = conditionString;
					if (conditionValue == "NULL") conditionValue = "";

					// Create the condition node using the stored values
					condition = new Condition("condition" + Conditions.ChildCount, this.Conditions);
					condition.Connector.Update(conditionConnector); condition.Key.Update(conditionKey);
					condition.Operator.Update(conditionOperator); condition.Value.Update(conditionValue);

					parseState = ParseState.Connector;
					break;
				}
			}
		}

		/// <summary> Returns a string that represents the current <see cref="XUIML.Data.Context"/>. </summary>
		/// <returns> A string that represents the current <see cref="XUIML.Data.Context"/>. </returns>
		public override string ToString ()
		{
			// The resulting string
			string result = "";

			// Serialize each condition
			for(int conditionIndex = 0; conditionIndex < Conditions.ChildCount; conditionIndex++)
			{
				// Get each condition
				Condition condition = Conditions[conditionIndex];

				// If its is not the first condition, add a connector
				if (conditionIndex > 0) result += condition.Connector + " ";

				// Serialize the condition into a string
				result += condition.Key + " " + condition.Operator + " " + condition.Value + " ";
			}

			// Return the resulting string
			return result;
		}

		#endregion

	}
}

