using UnityEngine;
using System;

namespace XUIML.Data
{
	/// <summary> Describes a Vector XUIML Node. </summary>
	public partial class Vector : Node
	{
		#region Public Properties
		
		/// <summary> The X coordinate value. </summary>
		public Number X { get; private set; }
		
		/// <summary> The Y coordinate value. </summary>
		public Number Y { get; private set; }

		/// <summary> The Z coordinate value. </summary>
		public Number Z { get; private set; }

		/// <summary> Gets a boolean value indicating whether the current values are default or not. </summary>
		public bool IsDefault { get { return (X.IsDefault && Y.IsDefault && Z.IsDefault); } }

		#endregion
		
		#region Public Constructor
		
		/// <summary> Initializes a new instance of the <see cref="XUIML.Vector"/> class. </summary>
		/// <param name="name"> The node name. </param>
		/// <param name="parent"> The parent node. </param>
		public Vector (string name, Node parent = null) : base(name, parent)
		{ this.X = new Number("x", this); this.Y = new Number("y", this); this.Z = new Number("z", this); }

		/// <summary> Initializes a new instance of the <see cref="XUIML.Vector"/> class. </summary>
		/// <param name="name"> The node name. </param>
		/// <param name="parent"> The parent node. </param>
		/// <param name="x"> The X coordinate value. </param>
		/// <param name="y"> The Y coordinate value. </param>
		/// <param name="z"> The Z coordinate value. </param>
		public Vector (string name, Node parent, float x, float y, float z) : base(name, parent)
		{ this.X = new Number("x", this, x); this.Y = new Number("y", this, y); this.Z = new Number("z", this, z); }

		#endregion

		#region Public Methods

		/// <summary> Update the node with specific values. </summary>
		///  <param name="vector"> The spatial values. </param>
		public void Update(Vector vector) { this.X.Value = vector.X; this.Y.Value = vector.Y; this.Z.Value = vector.Z;}

		/// <summary> Update the node with specific values. </summary>
		/// <param name="x"> The X vector component. </param>
		/// <param name="y"> The Y vector component. </param>
		/// <param name="z"> The Z vectorcomponent. </param>
		public void Update(Number x, Number y, Number z) { this.X.Value = x; this.Y.Value = y; this.Z.Value = z; }

		/// <summary> Update the node with specific values. </summary>
		/// <param name="x"> The X vector component. </param>
		/// <param name="y"> The Y vector component. </param>
		/// <param name="z"> The Z vectorcomponent. </param>
		public void Update(float x, float y, float z) { this.X.Value = x; this.Y.Value = y; this.Z.Value = z; }

		/// <summary> Obtains the object data from a string representation. </summary>
		/// <param name="value"> The string with the object data.</param>
		public override void FromString (string value) { Serialization.CSV.Deserialize(this, value); }

		/// <summary> Returns a string that represents the current object. </summary>
		/// <returns>A string that represents the current object.</returns>
		public override string ToString () { return string.Format ("[{0}, {1}, {2}]", this.X, this.Y, this.Z); }

		#endregion
	}
}
