using UnityEngine;
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Manages the Augmented Reality framework. </summary>
	public class AugmentedReality : MonoBehaviour
	{
		#region Public Subclasses

		/// <summary> Manages a Augmenter Reality marker. </summary>
		public sealed class Marker: ITrackableEventHandler
		{
			#region Public Properties

			/// <summary> Indicates whether the marker has been detected. </summary>
			public bool Detected { get; private set; }

			/// <summary> The entity object of the AR marker</summary>
			public GameObject EntityObject { get; private set; }

			public Text ResourceNode { get; private set; }

			#endregion

			#region Public Constructor

			/// <summary> Initializes a new instance of the <see cref="XUIML.Unity.AugmentedReality+Marker"/> class.
			/// </summary> <param name="gameObject"> The GameObject associated to the marker. </param>
			public Marker(GameObject gameObject)
			{
				this.EntityObject = gameObject;
				ImageTargetBehaviour behaviour = gameObject.GetComponent<ImageTargetBehaviour>();
				this.EntityObject.name = "Marker: " + behaviour.TrackableName;
				this.EntityObject.transform.parent = Manager.Instance.transform;
				behaviour.RegisterTrackableEventHandler(this);
			}

			#endregion

			#region ITrackableEventHandler implementation

			/// <summary> Handles the changes in the detection of the marker. </summary>
			/// <param name="previousStatus"> The previous detection status. </param>
			/// <param name="newStatus"> The new detection status. </param>
			public void OnTrackableStateChanged (TrackableBehaviour.Status previousStatus,
			                                     TrackableBehaviour.Status newStatus)
			{
				// Update the detection state
				Detected = (newStatus == TrackableBehaviour.Status.DETECTED ||
				            newStatus == TrackableBehaviour.Status.TRACKED ||
				            newStatus == TrackableBehaviour.Status.EXTENDED_TRACKED);
				
				//if (!Detected) CameraObject.transform.position = new Vector3 (0, 1000, 0); 

				if (ResourceNode != null) ResourceNode.Value = Detected.ToString();
				else if (Root.Instance != null) ResourceNode = Root.Instance.Platforms[0].Resources[0].Texts["MarkerDetected"];
			}

			#endregion
		}

		#endregion
		
		#region Public properties

		/// <summary> Indicates whether the manager has been initializated or not. </summary>
		public static bool Initialized { get; private set; }

		/// <summary> Indicates whether the manager is enabled or not. </summary>
		public static bool Enabled { get; set; }

		/// <summary> The dictionary of AR markers. </summary>
		public static Dictionary<string, Marker> Markers { get; private set; }

		/// <summary> The AR Camera object. </summary>
		public static GameObject CameraObject { get; private set; } 

		#endregion

		#region Private methods

		/// <summary> Initializes the instance. </summary>
		void Start()
		{
			// Mark the instance as not initialized
			Initialized = false; Enabled = true;

			// Initialize the Marker dictionary
			Markers = new Dictionary<string, Marker>();

			// TODO: Improve the creation of the AR camera object and Dataset loading
			AugmentedReality.CameraObject = Manager.Instantiate("Prefabs/ARCamera", "ARCamera", null);
			DataSetLoadBehaviour loader = AugmentedReality.CameraObject.GetComponent<DataSetLoadBehaviour>();
			loader.mDataSetsToLoad.Add("Workbench"); loader.mDataSetsToActivate.Add("Workbench");
		}

		/// <summary> Called on scene draw. </summary>
		void Update()
		{
			// Check if the AR Framework has been correctly initialized
			if(!Initialized)
			{
				// On play mode, make sure the AR gameobjecs are well organized
				if (UnityEngine.Application.isEditor)
				{
					if (GameObject.Find("TextureBufferCamera") == null) return;
					CameraObject.transform.parent = this.transform;
					GameObject.Find("TextureBufferCamera").transform.parent = this.transform;
					GameObject.Find("BackgroundCamera(Clone)").transform.parent = this.transform;
				}
				
				// Get an unidentified object
				GameObject unidentifiedMarker = GameObject.Find ("New Game Object");

				// Continue processing unidentified objects
				while (unidentifiedMarker != null)
				{
					// Extract the marker behaviour
					ImageTargetBehaviour markerData = unidentifiedMarker.GetComponent<ImageTargetBehaviour>();

					// Check if it really contains a marker
					if (markerData == null) continue;

					Markers.Add(markerData.TrackableName, new Marker(unidentifiedMarker));

					// Find the next possible marker
					unidentifiedMarker = GameObject.Find ("New Game Object");
				}

				// Mark the instance as initialized
				Initialized = true;
			}
			else // The AR framework has been initialized
			{
				if(!Enabled) return;

				// Check that there is at least one User
				if(Root.Instance.Users.ChildCount == 0) return;

				// Assign the user the point of view of the AR aamera
				User mainUser = Root.Instance.Users[0];
				Vector3 position = CameraObject.transform.position;
				Vector3 orientation = CameraObject.transform.eulerAngles;
				mainUser.Position.Update(position.x, position.y, position.z);
				mainUser.Orientation.Update(orientation.x, orientation.y, orientation.z);
				mainUser.FieldOfView.Update(CalculateFOV());
			}
		}

		/// <summary> Calculates the field of view from the AR camera. </summary>
		/// <returns> The field of view. </returns>
		/// <see cref="https://developer.vuforia.com/forum/unity-3-extension-technical-discussion/vertical-fov-unity"/>
		private float CalculateFOV()
		{ return 360.0f / Mathf.PI * Mathf.Atan(1.0f / CameraObject.camera.projectionMatrix[5]); }

		#endregion
	}
}