using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> A console to handle Debug data on mobile platform. </summary>
	public sealed class DebugConsole : MonoBehaviour
	{
		#region Private Subclasses

		/// <summary>  Encapsulates the debug data into a scrollable lists. </summary>
		private sealed class ScrollList
		{
			#region Public Subclasses

			/// <summary> Defines an element of a scroll list. </summary>
			public sealed class Element
			{
				/// <summary> The text of the element. </summary>
				public String Text;
				
				/// <summary> The text color of the element. </summary>
				public UnityEngine.Color Color;

				/// <summary> The level of the element.</summary>
				public int Level;

				/// <summary> Indicates whether the element is visible or not. </summary>
				public bool Visible;

				/// <summary> Indicates whether the child elements is visible or not. </summary>
				public bool Expanded;

				/// <summary> The node associated with the element. </summary>
				public Node Node;

				/// <summary> The child elements</summary>
				public List<Element> Children;
				
				/// <summary> The parent element reference. </summary>
				public Element Parent;
				
				/// <summary> Initializes a new instance of the <see cref="XUIML.DebugConsole+ScrollList+Element"/>
				/// class. </summary>
				/// <param name="text"> The text of the element. </param>
				/// <param name="color"> The text color. </param>
				/// <param name="parent"> The parent element reference. </param>
				/// <param name="expanded"> The visibility of the child elements. </param>
				/// <param name="node"> The node associated with the element. </param>
				public Element(string text, UnityEngine.Color color, Element parent, bool expanded, Node node)
				{
					// Initilize the public properties
					this.Text = text; this.Color = color; this.Parent = parent; this.Expanded = expanded;
					this.Node = node; this.Visible = true; this.Children = new List<Element>();
					
					// If there is a parent element, store its reference and add this instance to its children
					if (parent != null) { parent.Children.Add(this); this.Level = parent.Level + 1;} else Level = 0;
				}
			}
			#endregion
			
			#region Public Enumerations

			/// <summary> Enumerates the different kinds of buttons. </summary>
			public enum ButtonsData {Clear, Update};

			#endregion

			#region Public Properties

			/// <summary> The name of the scroll list. </summary>
			public string Name { get; private set; }

			/// <summary> The elements of the scroll list. </summary>
			public bool Updated { get; set; }

			/// <summary> The buttons of the scroll list. </summary>
			public ButtonsData Buttons { get; private set; }

			/// <summary> The current value of the scroll list. </summary>
			public Vector2 Scroll { get; private set; }

			/// <summary> The inertial value of the scroll list. </summary>
			public Vector2 ScrollInertia { get; private set; }

			/// <summary> The elements of the scroll list. </summary>
			public List<Element> Elements { get; private set; }

			#endregion

			#region Public Constructor

			/// <summary> Initializes a new instance of the <see cref="XUIML.DebugConsole+ScrollList"/> class. </summary>
			/// <param name="name"> The name of the scroll list. </param>
			/// <param name="buttons"> The buttons of the scroll list. </param>
			public ScrollList(string name, ButtonsData buttons)
			{ Name = name; Buttons = buttons; Scroll = new Vector2(); Updated = false; Elements = new List<Element>(); }

			#endregion

			#region Public Methods

			/// <summary> Adds the specified element to the list. </summary>
			/// <param name="element"> The element to add.</param>
			public void Add (Element element) { Elements.Add(element); }

			/// <summary> Clears the list of elements. </summary>
			public void Clear () { Elements.Clear(); }


			public void Draw(float width)
			{
				GUILayout.BeginVertical(GUILayout.Width (width));

				// Draw the title in the center
				GUILayout.BeginHorizontal(); GUILayout.Label(Name + ":"); GUILayout.EndHorizontal();

				// Start the scroll section
				Scroll = GUILayout.BeginScrollView(Scroll);

				// Go through each element text
				foreach (Element element in Elements)
				{	
					// Skip the invisible and non expanded elements (and mark their children)
					if (!element.Visible || (element.Parent != null && !element.Parent.Expanded))
					{ foreach (Element child in element.Children) child.Visible = false; continue;}
					else { foreach (Element child in element.Children) child.Visible = true; }

					// Draw the element
					GUI.contentColor = element.Color;
					GUILayout.BeginHorizontal();
					if (element.Level > 0) GUILayout.Space(16 * element.Level);
					if (element.Children.Count != 0)
					{
						if (element.Expanded) { if(GUILayout.Button("[--]")) element.Expanded = false; }
						else { if(GUILayout.Button("[+]")) element.Expanded = true; }
						GUILayout.Space(4);
					}
					if (GUILayout.Button(element.Text)) SelectedNode = element.Node;
					GUILayout.FlexibleSpace();
					GUILayout.EndHorizontal();
				}

				// Go back to white
				GUI.contentColor = White;
				
				// Finish the scroll section
				GUILayout.EndScrollView();

				Rect scrollArea = GUILayoutUtility.GetLastRect();
				if ((InputManager.CurrentAction == InputManager.Action.Move) && 
				    scrollArea.Contains(InputManager.RawPosition))
					ScrollInertia += (InputManager.RawPosition - InputManager.LastPosition) * Time.deltaTime;
					
				if (ScrollInertia.magnitude> 0.01f)
				{ Scroll += new Vector2(-ScrollInertia.x, ScrollInertia.y); ScrollInertia *= 0.9f; }

				GUILayout.BeginHorizontal();
				if (GUILayout.Button(new GUIContent("Clear"))) { this.Clear(); Updated = false; }
				GUILayout.EndHorizontal();

				GUILayout.EndVertical();
			}

			#endregion
		}

		#endregion

		#region Private Fields

		/// <summary> The size of the console message window. </summary>
		[NonSerialized] private Rect WindowSize = new Rect(20, 20, Screen.width - 40, Screen.height - 40);

		/// <summary> The console messages. </summary>
		private ScrollList ConsoleMessages;
		
		/// <summary> The entity hierarchy. </summary>
		private ScrollList EntityHierarchy;
		
		/// <summary> The list of attributes. </summary>
		private ScrollList EntityAttributes;

		/// <summary> The selected node. </summary>
		private static Node SelectedNode;

		/// <summary> The skin of the debug console. </summary>
		private GUISkin Skin;

		/// <summary> White color. </summary>
		private static UnityEngine.Color White = UnityEngine.Color.white;

		/// <summary> Yellow color. </summary>
		private static UnityEngine.Color Yellow = UnityEngine.Color.yellow;

		/// <summary> Red color. </summary>
		private static UnityEngine.Color Red = UnityEngine.Color.red;

		/// <summary> The key that togles the visibility of the debug console </summary>
		[SerializeField] private static KeyCode ToggleKey = KeyCode.Escape;
		
		/// <summary> Indicates whether to show the console window or not. </summary>
		[SerializeField] private static int ActivationCountdown = 3;

		#endregion

		#region Private Methods

		/// <summary> Called upon scene load. </summary>
		private void Awake ()
		{
			ConsoleMessages = new ScrollList("Console Messages", ScrollList.ButtonsData.Clear);
			EntityHierarchy = new ScrollList("Entity Hierarchy", ScrollList.ButtonsData.Update);
			EntityAttributes = new ScrollList("Entity Attributes", ScrollList.ButtonsData.Update);

			float fontSize = Screen.width / 70.0f;
			Skin = ScriptableObject.CreateInstance<GUISkin>();
			Skin.label.normal.textColor = White; Skin.label.fontSize = (int)(fontSize * 1.5f);
			Skin.button.normal.textColor = White; Skin.button.fontSize = (int)(fontSize * 1.2f);
			Skin.window.normal.textColor = White; Skin.window.fontSize = (int)(fontSize * 2.0f);
			Skin.window.border = new RectOffset(4,4,64,4);
		}

		/// <summary> Called upon scene rendering. </summary>
		private void Update()
		{ if(Input.GetKeyDown(ToggleKey)) { if(ActivationCountdown>0)ActivationCountdown--;else ActivationCountdown=1;}}

		/// <summary> Activates the console message log. </summary>
		private void OnEnable () { UnityEngine.Application.RegisterLogCallback(HandleConsoleMessages); }
		
		/// <summary> Deactivates the console message log. </summary>
		private void OnDisable () { UnityEngine.Application.RegisterLogCallback(null); }

		/// <summary> Called upon GUI drawing event. </summary>
		private void OnGUI ()
		{ if (ActivationCountdown == 0) WindowSize = GUILayout.Window(12, WindowSize, DrawWindow, "Debug Console"); }

		/// <summary> Displays a window with the logged console messages. </summary>
		/// <param name="windowID"> The window's ID.</param>
		private void DrawWindow (int windowID)
		{
			GUI.skin = this.Skin;

			HandleNodeHierarchy(null, null);
			//HandleObjectHierarchy(null, null);
			HandleEntityAttributes();
			GUILayout.BeginHorizontal();
//			ConsoleMessages.Draw(WindowSize.width / 1.1f);
			ConsoleMessages.Draw(WindowSize.width / 3.1f); GUILayout.FlexibleSpace();
			EntityHierarchy.Draw(WindowSize.width / 3.1f); GUILayout.FlexibleSpace();
			EntityAttributes.Draw(WindowSize.width / 3.1f);
			GUILayout.EndHorizontal();
			
			// Set the window to be draggable by the top title bar
			GUI.DragWindow(new Rect(0, 0, 10000, 40));
		}

		/// <summary> Handles the console message log. </summary>
		/// <param name="message"> The console message string. </param>
		/// <param name="stackTrace"> The console message stack trace. </param>
		/// <param name="type"> The console message type. </param>
		private void HandleConsoleMessages (string message, string stackTrace, LogType type)
		{ 
			switch (type) {
			case LogType.Log :

				// Filter out normal Vuforia messages
				if (message == "StartQCAR" || message == "StopQCAR" ||
					message.StartsWith("Can't add component") ||
				    message.StartsWith("QCARWrapper") ||
				    message.StartsWith("Creating Image") ||
				    message.StartsWith("VideoTextureInfo") ||
				    message.StartsWith("Setting camera clear flags") ||
				    message.StartsWith("SetHint")) 
					return;

				ConsoleMessages.Add(new ScrollList.Element(message, White, null, true, null));
				break;
			case LogType.Warning :
				ConsoleMessages.Add(new ScrollList.Element(message, Yellow, null, true, null));
				break;
			case LogType.Error :
				ConsoleMessages.Add(new ScrollList.Element(message + "\n" + stackTrace, Red, null, true, null));
				ActivationCountdown = 0; // Show the console on error
				break;
			}
		}

		private void HandleObjectHierarchy(Transform parentObject, ScrollList.Element parentElement)
		{
			// Check if the list is already updated
			if (EntityHierarchy.Updated) return;
			
			// Check if there is a parent node
			if (parentObject == null)
			{
				// Clear the list of objects
				EntityHierarchy.Clear();

				GameObject[] gameObjects = GameObject.FindObjectsOfType<GameObject>();

				// Travel through the root node children nodes
				foreach (GameObject go in gameObjects)
				{
					if (go.transform.parent != null) continue;

					// Create a new element
					ScrollList.Element newElement = new ScrollList.Element(go.name, White, null, (parentElement != null), null);
					
					// Add the element to the scroll list
					EntityHierarchy.Add(newElement);
					if (go.transform.parent == null) HandleObjectHierarchy(go.transform, newElement);
				}

				// Mark the scroll list as updated
				EntityHierarchy.Updated = true;
			}
			else 
			{
				// Travel through The child nodes
				foreach (Transform go in parentObject)
				{
					// Create a new element
					ScrollList.Element newElement = new ScrollList.Element(go.name, White, parentElement, (parentElement != null), null);
					
					// Add the element to the scroll list
					EntityHierarchy.Add(newElement);
					
					// Recursivelly handle the child gamobjects
					HandleObjectHierarchy(go, newElement);
				}
			}
		}

		private void HandleNodeHierarchy(Node parentNode, ScrollList.Element parentElement)
		{
			// Check if the list is already updated
			if (EntityHierarchy.Updated) return;

			// Check if there is a parent node
			if (parentNode == null)
			{
				// Clear the list of objects
				EntityHierarchy.Clear();

				// Travel through the root node children nodes
				if (Root.Instance != null) foreach (Node nodeset in Root.Instance) HandleNodeHierarchy(nodeset, null);

				// Mark the scroll list as updated
				EntityHierarchy.Updated = true;
			}
			else 
			{
				// Travel through The child nodes
				foreach (Node node in parentNode)
				{
					if (node.GetType().IsGenericType) { HandleNodeHierarchy(node, parentElement); continue; }

					// Skip nodes that are not represented as an entity
					if (!parentNode.GetType().IsGenericType) continue;

					bool expanded = false; //(parentElement == null) || (parentElement.Level > 2);

					string name = parentNode.Name.ToUpper() + ": " + node.Name;

					// Create a new element
					ScrollList.Element newElement = new ScrollList.Element(name, White, parentElement, expanded, node);

					if (!node.Updated) newElement.Color = Red;

					// Add the element to the scroll list
					EntityHierarchy.Add(newElement);
					
					// Recursivelly handle the child gamobjects
					HandleNodeHierarchy(node, newElement);
				}
			}
		}

		public void HandleEntityAttributes()
		{

			if (SelectedNode == null) return;

			EntityAttributes.Clear();

			if (SelectedNode is Text || SelectedNode is Number || SelectedNode is Index)
			{ EntityAttributes.Add(new ScrollList.Element(SelectedNode.ToString(), White, null, false, null)); return; }

			foreach (Node childNode in SelectedNode)
			{
				if (childNode.GetType().IsGenericType) continue;

				// Add the element to the scroll list
				EntityAttributes.Add(new ScrollList.Element(childNode.Name +": "+ childNode.ToString(), White,
				                                            null, false, childNode));
			}
		}


		#endregion

		#region Public Methods

		
		/// <summary> Indicates if the console is activated or not. </summary>
		public static bool IsActivated() { return (ActivationCountdown == 0); }

		/// <summary> Logs the specified debug message. </summary>
		/// <param name="message"> The debug message.</param>
		public static void Log(string message) { Debug.Log(message); }
		
		/// <summary> Logs the specified warning message. </summary>
		/// <param name="message"> The warning message.</param>	
		public static void LogWarning(string message) { Debug.LogWarning(message); }
		
		/// <summary> Clears the console window. </summary>
//		public static void Clear() { Instance.ConsoleMessages.Clear(); }

		public static void DrawStar(Vector3 position, float size = 0.1f, float duration = 1f, string colorName ="white")
		{
			UnityEngine.Color color = UnityEngine.Color.white;
			if (colorName == "red") color = UnityEngine.Color.red;
			if (colorName == "blue") color = UnityEngine.Color.blue;
			if (colorName == "green") color = UnityEngine.Color.green;
			Debug.DrawLine(position + new Vector3(-size/2, 0, 0), position + new Vector3(size/2, 0, 0), color, duration, true);
			Debug.DrawLine(position + new Vector3(0, -size/2, 0), position + new Vector3(0, size/2, 0), color, duration, true);
			Debug.DrawLine(position + new Vector3(0, 0, -size/2), position + new Vector3(0, 0, size/2), color, duration, true);
		}
		
		public static void DrawBoxCollider(BoxCollider collider, float duration = 1f, string colorName ="white")
		{
			UnityEngine.Color color = UnityEngine.Color.white;
			if (colorName == "red") color = UnityEngine.Color.red;
			if (colorName == "blue") color = UnityEngine.Color.blue;
			if (colorName == "green") color = UnityEngine.Color.green;
			
			Vector3 center = collider.center + collider.gameObject.transform.position;
			Vector3 size = collider.size;
			Vector3[] vertices = new Vector3[]
			{
				new Vector3(center.x - (size.x / 2), center.y - (size.y / 2), center.z - (size.z / 2)),
				new Vector3(center.x + (size.x / 2), center.y - (size.y / 2), center.z - (size.z / 2)),
				new Vector3(center.x + (size.x / 2), center.y + (size.y / 2), center.z - (size.z / 2)),
				new Vector3(center.x - (size.x / 2), center.y + (size.y / 2), center.z - (size.z / 2)),
				new Vector3(center.x - (size.x / 2), center.y - (size.y / 2), center.z + (size.z / 2)),
				new Vector3(center.x + (size.x / 2), center.y - (size.y / 2), center.z + (size.z / 2)),
				new Vector3(center.x + (size.x / 2), center.y + (size.y / 2), center.z + (size.z / 2)),
				new Vector3(center.x - (size.x / 2), center.y + (size.y / 2), center.z + (size.z / 2)),
			};
			
			Debug.DrawLine(vertices[0], vertices[1], color, duration, true);
			Debug.DrawLine(vertices[1], vertices[2], color, duration, true);
			Debug.DrawLine(vertices[2], vertices[3], color, duration, true);
			Debug.DrawLine(vertices[3], vertices[0], color, duration, true);
			Debug.DrawLine(vertices[4], vertices[5], color, duration, true);
			Debug.DrawLine(vertices[5], vertices[6], color, duration, true);
			Debug.DrawLine(vertices[6], vertices[7], color, duration, true);
			Debug.DrawLine(vertices[7], vertices[4], color, duration, true);
			Debug.DrawLine(vertices[0], vertices[4], color, duration, true);
			Debug.DrawLine(vertices[1], vertices[5], color, duration, true);
			Debug.DrawLine(vertices[2], vertices[6], color, duration, true);
			Debug.DrawLine(vertices[3], vertices[7], color, duration, true);
		}

		public static void DrawMeshCollider(MeshCollider collider, float duration = 1f, string colorName ="white")
		{
			UnityEngine.Color color = UnityEngine.Color.white;
			if (colorName == "red") color = UnityEngine.Color.red;
			if (colorName == "blue") color = UnityEngine.Color.blue;
			if (colorName == "green") color = UnityEngine.Color.green;

			UnityEngine.Mesh mesh = collider.sharedMesh;

			for (int i = 0; i < mesh.triangles.Length; i += 3)
			{
				Vector3[] triangleVertices = new Vector3[]
				{
					mesh.vertices[mesh.triangles[i+0]] + collider.gameObject.transform.position,
					mesh.vertices[mesh.triangles[i+1]] + collider.gameObject.transform.position,
					mesh.vertices[mesh.triangles[i+2]] + collider.gameObject.transform.position
				};

				Debug.DrawLine(triangleVertices[0], triangleVertices[1], color, duration, true);
				Debug.DrawLine(triangleVertices[1], triangleVertices[2], color, duration, true);
				Debug.DrawLine(triangleVertices[2], triangleVertices[0], color, duration, true);
			}
		}

#endregion
	}
}

