using System.Collections.Generic;
using UnityEngine;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Defines the representation of a XUIML Node in the Unity engine. </summary>
	public sealed class Element : Representation
	{
		#region Private Fields

		/// <summary> The number of layers currently active. </summary>
		private static int LayerCounter = 8;

		/// <summary> The default enviroment. </summary>
		private static Environment DefaultEnvironment;

		/// <summary> The current theme resource path. </summary>
		private static string CurrentThemePath = "Themes/Futuristic/";

		#endregion

		#region Public Properties

		/// <summary> The Unity GameObject that represents the Element. </summary>
		public GameObject Existence { get; private set; }

		/// <summary> The resource path to the GameObject. </summary>
		public string ResourcePath { get; private set; }

		/// <summary> The Entity node associated with this Element. </summary>
		public Entity Presence { get; private set; }

		/// <summary> The bounding box. </summary>
		public Bounds BoundingBox { get; private set;}

		/// <summary> The Unity Layer associated to the Element. </summary>
		public int Layer { get { return Existence.layer; } private set { SetLayer(value, Existence); } }

		/// <summary> The parent Element instance. </summary>
		public Element Parent { get; private set; }

		/// <summary> The child Element instances. </summary>
		public List<Element> Children { get; private set; }

		/// <summary> The Unity GameObject that allows a better integratio ninto the hierarchy. </summary>
		public GameObject ChildConnection { get; private set; }

		#endregion

		#region Private Constructor

		/// <summary> Initializes a new instance of the <see cref="XUIML.Unity.Element"/> class. </summary>
		/// <param name="node"> The XUIML node associated to the element. </param>
		private Element (Node node)
		{ this.Node = node; this.Name = node.Name; this.Children = new List<Element>(); }

		#endregion

		#region Private Methods

		/// <summary> Recursively sets the Unity layer value. </summary>
		/// <param name="layer"> The Unity layer value. </param>
		/// <param name="gameObject"> The GameObject instance to apply the layer value to. </param>
		private void SetLayer(int layer, GameObject gameObject)
		{
			// Make sure the GameObject exists
			if (gameObject == null) return;

			// Apply the layer value
			gameObject.layer = layer;

			// Avoid applying the layer to child connections
			if (gameObject == ChildConnection) return;

			// Apply the layer to the child GameObject instances
			foreach (Transform child in gameObject.transform) SetLayer(layer, child.gameObject);
		}

		/// <summary> Sets the parent Element reference, ensuring a propper connection. </summary>
		/// <param name="parent"> The parent Element. </param>
		private void SetParent(Element parent)
		{ 
			// Set the parent reference
			this.Parent = parent; this.Parent.Children.Add(this); this.Parent.Update();

			// Adjust the GameObject layer, bounding box and hierarchy position
			SetLayer(parent.Layer, this.Existence); BoundingBox = parent.BoundingBox;
			if (parent.ChildConnection != null) this.Existence.transform.parent = parent.ChildConnection.transform;
			else if (parent.Existence != null) this.Existence.transform.parent = parent.Existence.transform;
		}

		/// <summary> Gets the Unity representation of a given node. </summary>
		/// <returns> The Unity representation of the given node.</returns>
		/// <param name="node"> A XUIML node.</param>
		private Element GetElement(Node node)
		{
			// Check the given node
			if (node == null) return null;

			// Get the node representation
			Element element = (Element)node.Representation;

			// If there is no GameObject instance, update the representation
			if (element.Existence == null) element.Update();

			// Return the obtained representation
			return element;
		}

		/// <summary> Gets the Element component of a node.</summary>
		/// <returns>The element component.</returns>
		/// <param name="node"> The Node from which to extract the Element component.</param>
		private ElementComponent GetElementComponent<ElementComponent>(Node node) where ElementComponent : Component
		{ return GetElement(node).Existence.GetComponent<ElementComponent>(); }

		#endregion

		#region Public Methods

		/// <summary> Creates a new Unity representation for a given XUIML node. </summary>
		/// <param name="node"> The XUIML node. </param>
		public static void Create(Node node) { node.Representation = new Element(node); }

		/// <summary> Update this instance. </summary>
		public override void Update()
		{
			// Depending on the node type, perform different actions
			if (this.Node is Environment)
			{
				// Get the Environment node
				Environment environment = (Environment) this.Node;
				
				// Check if the GameObject instance already exists
				if (this.Existence == null)
				{
					// Create a void GameObject
					this.Existence = Manager.Instantiate(null, this.Name, null);
					
					// Asign a new layer number
					SetLayer(LayerCounter++, this.Existence);
					
					// Check if the current environment qualifies for being the default one
					if (DefaultEnvironment == null && environment.Type == "virtual") DefaultEnvironment = environment;
				}
			}
			else if (this.Node is Entity)
			{
				// Get the Entity node
				Entity entity = (Entity) this.Node;
				
				// Check if the GameObject instance already exists
				if (this.Existence == null)
				{
					// Create the GameObject instance
					this.Existence =  Manager.Instantiate(entity.Type, this.Name, null);

					// Get the parent Entity
					this.Parent = GetElement(this.Node.Ancestor<XUIML.Model.Entity>()); 

					// If no parent Entity has been found, use the Environment instance
					if (this.Parent == null)
					{
						Environment environment = entity.Ancestor<Environment>();
						if (environment == null) environment = DefaultEnvironment;
						this.Parent = GetElement(environment);
					}

					// Adjust the hierarchy
					SetParent(this.Parent);
				}

				// Adjust the transform of the GameObject
				this.Existence.transform.localPosition = ToVector3(entity.Position);
				this.Existence.transform.localEulerAngles = ToVector3(entity.Rotation);

			}

			else if (this.Node is Force)
			{
				// Get the Entity node
				Force force = (Force) this.Node;
				
				// Check if the GameObject instance already exists
				if (this.Existence == null)
				{
					// Create the GameObject instance
					this.Existence = new GameObject(force.Name);
					
					// Get the parent Entity
					this.Parent = GetElement(this.Node.Ancestor<XUIML.Model.Entity>()); 
					
					// If no parent Entity has been found, use the Environment instance
					if (this.Parent == null)
					{
						Environment environment = force.Ancestor<Environment>();
						if (environment == null) environment = DefaultEnvironment;
						this.Parent = GetElement(environment);
					}
					
					// Adjust the hierarchy
					SetParent(this.Parent);
				}
				
				// Adjust the transform of the GameObject
				this.Existence.transform.localPosition = ToVector3(force.Position);
				this.Existence.transform.localEulerAngles = ToVector3(force.Rotation);

				// If the force is electromagnetic, convert it to a light
				if (force.Effect == "electromagnetic")
				{
					// Check if it is ambient light
					if (force.Position.IsDefault && force.Rotation.IsDefault)
						RenderSettings.ambientLight = ToColor(force.Color) * force.Magnitude;
					else // Otherwise, is a spatial light 
					{
						// Create a new light component
						Light light = this.Existence.AddComponent<Light>();

						// Set the right type of light
						if (force.Position.IsDefault) light.type = LightType.Directional;
						else if (force.Angle.IsDefault) light.type = LightType.Point;
						else { light.type = LightType.Spot; light.spotAngle = force.Angle; }

						// Assign the node values
						light.intensity = force.Magnitude;
						light.color = ToColor(force.Color);
						light.shadows = LightShadows.None;//Hard;
						light.shadowBias = 0.02f;
					}
				}
			}
			else if (this.Node is User)
			{
				// Get the User node
				User user = (User) this.Node;

				// Check if the child entities instance already exists
				if (this.Children.Count != Root.Instance.Environments.ChildCount)
				{
					foreach (Environment environment in Root.Instance.Environments)
					{
						if (!environment.Entities.Contains("User Presence: " + this.Name))
						{
							// Create a new entity to represent the user
							Entity presence = new Entity("User Presence: " + this.Name, environment.Entities); 
							Element element = GetElement(presence);

							// Set the initial camera properties
							GameObject cameraObject = element.Existence;
							cameraObject.AddComponent<Camera>();
							cameraObject.camera.cullingMask = (int)Mathf.Pow(2, element.Layer);
							cameraObject.camera.clearFlags = CameraClearFlags.Depth;
							cameraObject.camera.nearClipPlane = 0.01f;
							cameraObject.camera.farClipPlane = 100.0f;
							presence.Position.Update(0f, 0.33f, 0f);
							presence.Rotation.Update(90f, 0f, 0f);

							// Add the representation to the child list
							Children.Add(element);
						}
					}
				}

				// Update the user presence entities
				foreach(Element element in this.Children)
				{
					// Get the associated Entity and Environment nodes
					Entity presence = (Entity)element.Node;
					Environment environment = presence.Ancestor<Environment>();

					// If required, update the user presence in the environment
					if (environment.Type == "physical" || !environment.Link.IsEmpty)
					{ presence.Position.Update(user.Position); presence.Rotation.Update(user.Orientation); }

					// Update the field of view
					if (!user.FieldOfView.Updated) GetElement(presence).Existence.camera.fieldOfView = user.FieldOfView;
				}
			}

			else if (this.Node is Context)
			{
				// Get the Context node and its parent
				Context context = (Context) this.Node;
				Widget parentWidget = context.Ancestor<Widget>();

				// Ignore void contexts
				if (context.Conditions.ChildCount == 0) return;

				// Check the parent contexts
				// TODO: Improve this solution (reorder the conditions)
				if (context.Conditions[0].Node == null)
				{
					Widget searchNode = parentWidget.Ancestor<Widget>();
					while (searchNode != null)
					{
						foreach (Context.Condition condition in searchNode.Context.Conditions)
						{
							if (!context.Contains(condition.Key))
							{
								Context.Condition newCondition = new Context.Condition
									("condition" + context.Conditions.ChildCount, context.Conditions);
								newCondition.Connector.Value = condition.Connector.Value;
								newCondition.Key.Value = condition.Key.Value;
								newCondition.Operator.Value = condition.Operator.Value;
								newCondition.Value.Value = condition.Value.Value;
							}
						}

						searchNode = searchNode.Ancestor<Widget>();
					}
					context.Integrated = true;
				}

				// Evaluate all the conditions of the context
				context.Active = true;
				foreach (Context.Condition condition in context.Conditions)
				{
					if (condition.Node == null)
					{
						condition.Node = Manager.GetResource(null, condition.Key);
						if (condition.Node == null)	condition.Node = Manager.GetResource(parentWidget, condition.Key);
						if (condition.Node == null)	throw new System.Exception
							("Invalid condition node: " + condition.Key + " for " + this.Node.Parent.Name);

						// Add an event connection to the 
						condition.Node.UpdatePerformed += condition.RequestUpdate;
					}

					// The current condition result
					condition.Active = false;

					// First, try to analyze the condition using strings
					string string1 = ((Text)condition.Node).Value; string string2 = condition.Value.Value;
					if (condition.Operator == "=") condition.Active = (string1 == string2);
					else if (condition.Operator == "!=") condition.Active = (string1 != string2);
					else
					{
						// if that does not work, analyze it as numbers
						float number1 = float.Parse(string1); float number2 = float.Parse(string2);
						if (condition.Operator == ">") condition.Active = (number1 > number2);
						else if (condition.Operator == "<") condition.Active = (number1 < number2);
						else if (condition.Operator == ">=") condition.Active = (number1 >= number2);
						else if (condition.Operator == "<=") condition.Active = (number1 <= number2);
					}

					// Use the appropiate boolean connector
					if (condition.Connector == "and") context.Active = (context.Active && condition.Active);
					else if (condition.Connector == "or") context.Active = (context.Active || condition.Active);
				}
				
				// Adjust the visibility of the presences
				GetElementComponent<WidgetBehaviour>(parentWidget).Enabled = context.Active;
			}
			else if (this.Node is Widget)
			{
				// Get the Widget node and its parent
				Widget widget = (Widget) this.Node;
				Widget parentWidget = widget.Ancestor<Widget>();
				WidgetBehaviour behaviour;

				// Check if the GameObject instance already exists
				if (this.Presence == null)
				{
					// Get the parent node
					Node parentNode;
					
					// Check if the widget has an environment assigned
					if (!widget.Environment.IsEmpty)
					{
						// Set the environment has the parent node
						parentNode = Root.Instance.Environments[widget.Environment].Entities;
						this.Parent = GetElement(Root.Instance.Environments[widget.Environment]);
						parentWidget = null;
					}
					else 
					{
						// If there is no parent widget, use the default environment
						if (parentWidget == null)
						{ this.Parent = GetElement(DefaultEnvironment); parentNode = DefaultEnvironment.Entities; }
						else 
						{ parentWidget.Update(); Parent=GetElement(parentWidget); parentNode=Parent.Presence.Entities; }
					}

					// Initialize the widget instace
					Presence = new Entity(widget.Name, parentNode);
					Presence.Type.Value = CurrentThemePath + "Widgets/" + widget.Type;
					Presence.Update();
					this.Existence = GetElement(Presence).Existence;
					this.Parent.Children.Add(this);
					BoundingBox = new Bounds(Vector3.zero,  Parent.BoundingBox.size);
					behaviour = this.Existence.GetComponent<WidgetBehaviour>();
					if (behaviour == null) behaviour = this.Existence.AddComponent<WidgetBehaviour>();
					behaviour.Initialize(widget, this.Parent.Existence.GetComponent<WidgetBehaviour>());

					if (parentWidget != null && behaviour.Parent != null)
					{
						GameObject connector = this.Parent.Existence.GetComponent<WidgetBehaviour>().Connector;
						if (connector != null) { Existence.transform.parent = connector.transform; Presence.Update(); }
					}

					if (!widget.Location.IsEmpty)
					{
						// Look for the first widget
						XUIML.Model.Environment environment = DefaultEnvironment;
						Widget searchWidget = widget;
						while (searchWidget != null && searchWidget.Environment.IsEmpty)
							searchWidget = searchWidget.Ancestor<Widget>();
						if (searchWidget != null) environment  = Root.Instance.Environments[searchWidget.Environment];
						
						Location location = environment.Locations[widget.Location];
						if (location == null) throw new System.Exception("Invalid Location");
						
						Vector3 center = ToVector3(location.Position); Vector3 size = ToVector3(location.Size);
						this.BoundingBox = new Bounds(center, size);

						widget.Rotation.Update(location.Rotation);
					}

					if (!widget.Size.IsDefault) BoundingBox = new Bounds(BoundingBox.center, ToVector3(widget.Size));

					if (widget.Location.IsEmpty && parentWidget != null && parentWidget.Widgets.ChildCount > 1
						&& BoundingBox.size.magnitude != 0)
					{
						float totalIndex = parentWidget.Widgets.ChildCount;
						float currentIndex = parentWidget.Widgets.IndexOf(widget);
						Vector3 center = BoundingBox.center, size = BoundingBox.size, min = BoundingBox.min;

						if (size.x >= size.z) { size.x /= totalIndex; center.x = min.x+((currentIndex+0.5f) * size.x); }
						else {size.z /= totalIndex; center.z -= min.z + ((currentIndex + 0.5f) * size.z);}

						this.BoundingBox = new Bounds(center, size);
					}

					// Create a connection for the text nodes
					foreach (TextMesh textObject in behaviour.Texts)
					{
						Text text = Manager.GetResource(widget, textObject.name);
						if (text != null) text.UpdatePerformed += widget.RequestUpdate;
						if (!behaviour.TextNodes.Contains(text)) behaviour.TextNodes.Add(text);
					}

					// Create a connection for the images
					foreach (Image imageObject in behaviour.Images)
					{
						if (imageObject.FileName == null) continue;
						Text imagePath = Manager.GetResource(widget, imageObject.FileName);
						imageObject.gameObject.renderer.material.mainTexture = Resources.Load(imagePath.Value) as Texture;
					}
				}

				behaviour = this.Existence.GetComponent<WidgetBehaviour>();

				// Update the position value
				if (widget.Position.IsDefault) Presence.Position.Update(ToVector(BoundingBox.center));
				else this.Presence.Position.Update(widget.Position);

				// Update the rotation value
				Presence.Rotation.Update(widget.Rotation);

				// Update the scale value value
				if (behaviour.Resizeable && Existence != null)
					foreach (Transform childObject in Existence.transform)
						childObject.localScale = ToVector3(widget.Size) * 10f;

				if (behaviour.Resizeable && Existence != null && Existence.collider != null)
					Existence.transform.localScale = ToVector3(widget.Size) * 10f;

				// Update the text elements
				foreach (TextMesh textObject in this.Existence.GetComponent<WidgetBehaviour>().Texts)
				{
					Text textNode = Manager.GetResource(widget, textObject.name);
					if (textNode != null) 
					{
						// Accept new line escape characters
						string text = textNode.Value.Replace("\\n","\n");

						// Update the text value
						textObject.text = text;

						// Restrict the size of the displayed texts
//						if(behaviour != null && behaviour.MaxTextSize > 0 && text.Length > behaviour.MaxTextSize)
//							text = "..." + text.Substring(text.Length - behaviour.MaxTextSize, behaviour.MaxTextSize);

						if(behaviour != null && behaviour.MaxTextSize > 0)
						{
							int displayedChars = text.Length;

							while (textObject.renderer.bounds.size.x > behaviour.MaxTextSize)
							{ textObject.text = "..." + text.Substring(text.Length-displayedChars, displayedChars--); }
						}
					}
				}

				// Update the Entity object
				this.Presence.Update();
			}
		}
		
		/// <summary> Destroys this instance. </summary>
		public override void Destroy()
		{
			foreach(Element child in Children) child.Destroy();
			
			if (Presence != null) Presence.Destroy();
			if (Existence != null) GameObject.Destroy(Existence);
		}

		/// <summary> Transforms a Vector node into a Vector3 instance. </summary>
		/// <param name="vector"> The Vector node. </param>
		/// <returns> The resulting Vector3 instance. </returns>
		public static Vector3 ToVector3(Vector vector)
		{ return new Vector3(vector.X, vector.Y, vector.Z); }
		
		/// <summary> Transforms a Vector3 instace into a Vector node. </summary>
		/// <param name="vector"> The Vector3 instace. </param>
		/// <returns> The Vector node. </returns>
		public static Vector ToVector(Vector3 vector)
		{ return new Vector("vector", null, vector.x, vector.y, vector.z); }
		
		/// <summary> Transforms a Color node into a Unity Color instance. </summary>
		/// <param name="color"> The Color node. </param>
		/// <returns> The resulting Unity Color instance. </returns>
		public static UnityEngine.Color ToColor(XUIML.Data.Color color)
		{ return new UnityEngine.Color(color.R, color.G, color.B); }
		
		/// <summary> Updates the scene materials. </summary>
		/// <param name="go"> The gameObject Instance whose mateerials should be updated </param>
		public static void UpdateMaterials(GameObject go = null)
		{
			// Populate the list of materials that can be colorized
			Dictionary<string, float> Materials = new Dictionary<string, float>();
			Materials.Add("Ground", 1.0f);
			Materials.Add("LogoMain", 0.8f);
			Materials.Add("ButtonSelected", 0.8f);
			Materials.Add("ButtonUnSelected", 0.4f);
			Materials.Add("KeyboardForeground", 0.8f);
			Materials.Add("KeyboardBackground", 0.4f);
			Materials.Add("KeySelected", 0.8f);
			Materials.Add("KeyUnselected", 0.4f);
			Materials.Add("WindowForeground", 0.8f);
			Materials.Add("WindowBackground", 0.4f);
			Materials.Add("ManipulationAchieved", 0.8f);
			Materials.Add("SelectionStructure", 0.8f);
			Materials.Add("Avatar", 0.8f);
			
			XUIML.Data.Color mainColor = Root.Instance.Users[0].Preferences[0].MainColor;
			
			// A (somewhat) dirty solution to change the color of the user interface
			Renderer[] renderers;
			if (go != null) renderers = go.GetComponentsInChildren<Renderer>();
			else renderers = GameObject.FindObjectsOfType<Renderer>();
			
			foreach (Renderer renderer in renderers)
			{
				foreach (string materialName in Materials.Keys)
				{
					if (renderer.material.name.Contains(materialName)) 
					{
						Vector2 offset = renderer.material.mainTextureOffset;
						renderer.material.color = new UnityEngine.Color(
							mainColor.R * Materials[materialName], 
							mainColor.G * Materials[materialName],
							mainColor.B * Materials[materialName],
							renderer.material.color.a);
						renderer.material.mainTextureOffset = offset;
					}
				}
			}
		}

		#endregion
	}
}