﻿using UnityEngine;
using System.Collections;

namespace XUIML.Unity
{
	public class Key : MonoBehaviour
	{
		#region Private Constants

		/// <summary> The time between presses. </summary>
		private const float TimeBetweenPresses = 0.5f;

		#endregion

		#region Private Fields

		/// <summary> The keyboard the Key belongs to. </summary>
		private Keyboard Keyboard;

		/// <summary> The GameObject with the key symbol. </summary>
		private GameObject Symbol;

		/// <summary> The key symbol TextMesh. </summary>
		private TextMesh SymbolText;

		/// <summary> The last key pressed. </summary>
		private static Key LastKeyPressed;

		/// <summary> The time since the last key was pressed. </summary>
		private static float LastTimePressed;
		

		#endregion

		#region Public Fields

		/// <summary> The current symbol value index. </summary>
		public int CurrentIndex;

		/// <summary> The symbol values.</summary>
		public string[] Values;

		/// <summary> The symbol font. </summary>
		public Font SymbolFont;

		/// <summary> The size of the symbol. </summary>
		public float SymbolSize;

		/// <summary> The material of the symbol. </summary>
		public Material SymbolMaterial;

		#endregion
		
		#region Public Methods

		/// <summary> Initializes the Key instance.  </summary>
		public void Awake()
		{
			// Get the parent keyboard reference
			Keyboard = GetComponentInParent<Keyboard>();

			// Create a Gameobject to present the symbol
			Symbol = new GameObject("Symbol") ;
			SymbolText = Symbol.AddComponent<TextMesh>();
			Symbol.transform.parent = this.transform;
			Symbol.layer = this.gameObject.layer;
			Symbol.transform.localPosition = new Vector3(0.0f, 0.0005f, 0.0005f);
			Symbol.transform.localEulerAngles = new Vector3(90.0f, 0.0f, 0.0f);
			if (Values.Length > 0) SymbolText.text = Values[0];

			// Se the text mesh properties
			SymbolText.font = SymbolFont;
			SymbolText.anchor = TextAnchor.MiddleCenter;
			SymbolText.alignment = TextAlignment.Center;
			SymbolText.fontSize = 64;
			SymbolText.characterSize = SymbolSize;
			Symbol.renderer.material = SymbolMaterial;
		}

		/// <summary> Called when the user presses on the key. </summary>
		/// <param name="hit"> The hit data. </param>
		public void OnPressed(RaycastHit hit)	{ LastTimePressed = 0; Press();}
		
		
		/// <summary> Called when the user continues pressing on the key. </summary>
		/// <param name="hit"> The hit data. </param>
		public void OnCrossed(RaycastHit hit)
		{
			if (LastKeyPressed == this)
			{
				LastTimePressed += Time.deltaTime;
				if (LastTimePressed > TimeBetweenPresses) {LastTimePressed = 0; Press();}
			}
			else { LastKeyPressed = this; LastTimePressed = 0; }
		}

		/// <summary> performs the actions associated with the key. </summary>
		public void Press()
		{
			// Play the "Click" sound
			OutputManager.PlaySound("Click", this.transform.position, 0.5f);

			// Send the signal to the keyboard
			Keyboard.KeyPressed(this); 
			
			// Play the appropiate animation
			if (animation) { animation.Stop(); animation.Play("HideKey"); animation.PlayQueued("ShowKey"); }
		}


		/// <summary> Updates a symbol value. </summary>
		/// <param name="index"> The symbol value index.</param>
		/// <param name="value"> The new symbol value.</param>
		public void UpdateValue(int index, string value)
		{ Values[index] = value; if (index == CurrentIndex) SymbolText.text = Values[index]; }


		/// <summary> Changes the symbol index. </summary>
		/// <param name="newIndex"> The new index.</param>
		public void ChangeIndex(int newIndex, bool forced = false)
		{
			// Create some local variables
			string currentValue = null, newValue = null;

			// Check the given value index
			if (Values.Length >= CurrentIndex) currentValue = Values[CurrentIndex];
			if (Values.Length >= newIndex) newValue = Values[newIndex];

			// Update the symbol
			if (currentValue != newValue || forced)
			{ 
				// Animate the transition
				animation.Stop();
				if (currentValue != null && currentValue != "") animation.PlayQueued("HideKey");
				if (newValue != null && newValue != "") animation.PlayQueued("ShowKey");

				// Update the text
				if (newValue != null) SymbolText.text = newValue; else SymbolText.text = "";
			}

			// Set the new index value
			CurrentIndex = newIndex;
		}

		#endregion
	}
}