using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> The abstract class that serves as the base for all (virtual/soft) Keyboards. </summary>
	public abstract class Keyboard : MonoBehaviour
	{
		#region Private Constants

		/// <summary> The scale factor multiplier value. </summary>
		private const float ScaleFactorMultiplier = 2.0f;

		/// <summary> The maximum scale factor value. </summary>
		private const float MinScaleFactor = 1.0f;

		/// <summary> The maximum scale factor value. </summary>
		private const float MaxScaleFactor = 1.75f;

		#endregion

		#region Protected Fields

		/// <summary> The keys of the keyboard. </summary>
		protected Key[] Keys;
		
		/// <summary> The text node associated to the keyboard. </summary>
		protected XUIML.Data.Text TextNode;
		
		/// <summary> The parent widget behaviour. </summary>
		protected WidgetBehaviour ParentWidget;
		
		/// <summary> The parent widget node. </summary>
		protected Widget ParentWidgetNode;

		/// <summary> The text meshes. </summary>
		protected Dictionary<string, TextMesh> Texts;

		#endregion

		#region Protected Methods

		/// <summary> Initializes the basic elements of the keyboard. </summary>
		protected void Initialize()
		{
			// Get the parent widget behaviour
			ParentWidget = transform.parent.parent.GetComponent<WidgetBehaviour>();
			ParentWidgetNode = ParentWidget.WidgetNode;
			
			// Get the text node associated to this keyboard
			if (ParentWidget != null && ParentWidget.TextNodes.Count != 0)TextNode = ParentWidget.TextNodes[1];

			// Get the keys of the keyboard
			Keys = GetComponentsInChildren<Key>();

			// The different textmesh components
			TextMesh[] textMeshes = GetComponentsInChildren<TextMesh>();
			Texts = new Dictionary<string, TextMesh>();
			foreach (TextMesh textMesh in textMeshes) if (textMesh.name != "Symbol") Texts.Add(textMesh.name, textMesh);
		}

		/// <summary> Updates this instance once per frame. </summary>
		protected virtual void Update()
		{
			// If the parent widget is not selected
			if (!ParentWidget.Selected) return;
			
			// Update the scale value
			float scaleFactor = Vector3.Distance(transform.position, AugmentedReality.CameraObject.transform.position)
			 					* ScaleFactorMultiplier;
			if (scaleFactor < MinScaleFactor) scaleFactor = MinScaleFactor; 
			else if (scaleFactor > MaxScaleFactor) scaleFactor = MaxScaleFactor;
			transform.localScale = new Vector3(scaleFactor, scaleFactor, scaleFactor);
		}

		#endregion

		#region Public Methods

		/// <summary> Called when a key is pressed. </summary>
		/// <param name="key"> The Key pressed.</param>
		public abstract void KeyPressed(Key key);

		#endregion
	}
}

