﻿using UnityEngine;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	///<summary> Handles the user input. </summary>
	public class InputManager : MonoBehaviour
	{
		#region Public Enumerations
		
		public enum Action { None, Press, Move, Release };
		
		#endregion

		#region Public Subclasses

		/// <summary> Defines a primary target for the input manager. </summary>
		public abstract class PrimaryTarget : MonoBehaviour {}

		#endregion
		
		#region Private Fields
		
		/// <summary> The current scene cameras. </summary>
		private List<Camera> CameraList;
		
		/// <summary> dont require receiver.
		/// </summary>
		private const SendMessageOptions ReceiverNotRequired = SendMessageOptions.DontRequireReceiver;
		
		#endregion
		
		#region Public Properties
		
		/// <summary> The current action. </summary>
		public static Action CurrentAction { get; private set; }
		
		/// <summary> Indicates whether the user is pressing a button or the sceen. </summary>
		public static bool Pressed { get; private set; }
		
		/// <summary> Indicates whether the ruser is pressing on the. </summary>
		public static bool Stationary { get; private set; }
		
		/// <summary> Stores the last hit information </summary>
		public static RaycastHit Hit { get; private set; }
		
		/// <summary> Stores the last hit position </summary>
		public static Vector3 HitPosition { get; private set; }
		
		/// <summary> The time since the last action. </summary>
		public static float TimeReleased;
		
		/// <summary> The time since the object was released. </summary>
		public static float TimePressed;

		/// <summary> The raw cursor position. </summary>
		public static Vector2 RawPosition;

		/// <summary> The last cursor position. </summary>
		public static Vector2 LastPosition;

		/// <summary> The initial cursor position. </summary>
		public static Vector2 InitialPosition;
		
		/// <summary> The initial GameObject. </summary>
		public static GameObject InitialObject;

		#endregion
		
		#region Private methods
		
		/// <summary> Initializes the instance. </summary>
		private void Start ()
		{
			// Get the list of camera objects
			CameraList = new List<Camera>();
			Root.Instance.Update();
			foreach (Environment environment in Root.Instance.Environments)
				foreach (Entity entity in environment.Entities) 
					if (entity.Name.StartsWith("User Presence")) 
						CameraList.Add(((Unity.Element)entity.Representation).Existence.camera);
			
			// Create a variable to store the last hit data
			Hit = new RaycastHit();
			
			// Define the timers and the associated nodes
			TimeReleased = TimePressed = 0f;
		}
		
		/// <summary> Called on drawing. </summary>
		private void Update ()
		{
			// Reset the variables
			CurrentAction = Action.None; Stationary = false; LastPosition = RawPosition;
			
			#if UNITY_EDITOR
			
			if (Input.GetMouseButtonDown(0)) CurrentAction = Action.Press;
			else if (Input.GetMouseButton(0)) CurrentAction = Action.Move;
			else if (Input.GetMouseButtonUp(0)) CurrentAction = Action.Release;
			
			if (RawPosition == (Vector2)Input.mousePosition) Stationary = true;
			RawPosition = (Vector2) Input.mousePosition;
			
			#else
			
			if (Input.touchCount == 0) return;
			
			Touch touch = Input.GetTouch(0);
			if (touch.phase.Equals(TouchPhase.Began)) CurrentAction = Action.Press;
			else if (touch.phase.Equals(TouchPhase.Stationary)) CurrentAction = Action.Move;
			else if (touch.phase.Equals(TouchPhase.Moved)) CurrentAction = Action.Move;
			else if (touch.phase.Equals(TouchPhase.Ended)) CurrentAction = Action.Release;
			else if (touch.phase.Equals(TouchPhase.Canceled)) CurrentAction = Action.Release;
			
			if (touch.phase.Equals(TouchPhase.Stationary)) Stationary = true;
			RawPosition = touch.position;
			
			#endif
			
			// If no actions have been performed (or the debug console is active), just update the timers and leave
			if (CurrentAction == Action.None || DebugConsole.IsActivated())
			{ TimeReleased +=Time.deltaTime; TimePressed = 0f; return; }
			
			// Update the timers
			TimeReleased +=Time.deltaTime;
			TimePressed = 0f;
			
			// Check all cameras (and layers)
			foreach (Camera cameraObject in CameraList)
			{
				// Calculate a ray from the current position
				Ray ray = cameraObject.ScreenPointToRay (RawPosition);
				
				int layerMask =(int) Mathf.Pow(2, cameraObject.gameObject.layer);
				bool primaryTargetDetected = false;
				RaycastHit currentHit = new RaycastHit(); currentHit.distance = Mathf.Infinity;
				RaycastHit[] hits = Physics.RaycastAll(ray, Mathf.Infinity, layerMask);
				foreach (RaycastHit hit in hits)
				{
					// Check if the current hit object is a primary target
					bool isPrimaryTarget = (hit.collider.gameObject.GetComponent<PrimaryTarget>() != null);
					
					// Check if a primary target has been detected before
					if (primaryTargetDetected)
					{
						// Only accept new manipulators that are closer to the user POV
						if (isPrimaryTarget && hit.distance < currentHit.distance) currentHit = hit;
					}
					else 
					{
						// If an object is a primary target, it does not have to pass the depth test
						if (isPrimaryTarget) { primaryTargetDetected = true; currentHit = hit;}
						else if (hit.distance < currentHit.distance) currentHit = hit;
					}
				}
				
				// Exclude those layers without a hit object
				if (currentHit.collider == null) continue;
				
				// Store the current hit
				Hit = currentHit;
				HitPosition = currentHit.point;
				
				#if UNITY_EDITOR

				// Unity Debug
				if (primaryTargetDetected) DebugConsole.DrawStar(Hit.point, 0.01f, 1, "red");
				else DebugConsole.DrawStar(Hit.point, 0.01f, 1, "white");

				//if (CurrentAction == Action.Press) Debug.LogWarning("Pressed on " + Hit.transform.name); 
				#endif
				
				GameObject hitObject = Hit.transform.gameObject;
				
				switch (CurrentAction)
				{
				case Action.Press:
					hitObject.SendMessage("OnPressed", Hit, ReceiverNotRequired);
					InitialObject = hitObject;
					break;
				case Action.Move:
					hitObject.SendMessage("OnCrossed", Hit, ReceiverNotRequired);
					if (InitialObject == hitObject) hitObject.SendMessage("OnDragged", Hit, ReceiverNotRequired);
					break;
				case Action.Release:
					hitObject.SendMessage("OnReleased", Hit, ReceiverNotRequired);
					if (InitialObject == hitObject) hitObject.SendMessage("OnClicked", Hit, ReceiverNotRequired);
					break;
				}
			}
			
			// Add a failsafe in case, the release operation occurs outside the initial object
			if (CurrentAction == Action.Release && InitialObject != null)
			{ InitialObject.SendMessage("OnReleasedWithoutPressing", Hit, ReceiverNotRequired); ;}
		}
		
		#endregion
		
		#region Private methods
		
		/// <summary> Records a sound using the inbuilt Microphone. </summary>
		public static void RecordSound (string name, int length)
		{
			// Record the sound
			AudioClip recordedSound = Microphone.Start(Microphone.devices[0], false, length, 11025);

			// Store the sound for playback
			OutputManager.AddSound(name, recordedSound); 
		}

		#endregion
	}
}

