using UnityEngine;
using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Manages the XUIML Unity implementation. </summary>
	public class Manager : MonoBehaviour
	{
		#region Public Fields

		/// <summary> Exposes the Root Document path to Unity editor for an easy access. </summary>
		public string RootDocumentPath;

		#endregion
		
		#region Public Properties

		/// <summary> Gets the XUIML DataSet root document. </summary>
		public static Document RootDocument { get; private set; }

		/// <summary> The singleton instance. </summary>
		public static Manager Instance { get; private set; }

		#endregion
		
		#region Private methods

		/// <summary> Initializes the instance. </summary>
		void Awake ()
		{
			// Make sure that the singleton pattern is correclty mantained
			if (Instance != null) throw new Exception("Only one XUIML Manager instance allowed."); else Instance = this;

			// Disable all other GameObjects in the current scene
			GameObject[] externalGameObjects = GameObject.FindObjectsOfType<GameObject>();
			foreach(GameObject externalGameObject in externalGameObjects)
				if (externalGameObject != this.gameObject) GameObject.DestroyImmediate(externalGameObject);

			Node.Created += Unity.Element.Create;

			// Create a new XUIML dataset and load the root document
			RootDocument = new Document(RootDocumentPath, new Root()); RootDocument.Deserialize();
		}

		/// <summary> Called on scene draw. </summary>
		void LateUpdate()
		{
			// All the "magic", the entire main loop, is a single line
			while (!Root.Instance.Updated) Root.Instance.Update();
		}

		/// <summary> Hides the navigation bar. </summary>
		private void HideNavigationBar()
		{
			#if UNITY_ANDROID && !UNITY_EDITOR
			using (var unityPlayer = new AndroidJavaClass("com.unity3d.player.UnityPlayer"))
			{	using (var activity = unityPlayer.GetStatic<AndroidJavaObject>("currentActivity"))
				{ activity.Call ("runOnUiThread", new AndroidJavaRunnable (HideNavigationBarRunable));	}
			}
			#endif
		}

		/// <summary> (Really) hides the navigation bar. </summary>
		public static void HideNavigationBarRunable ()
		{
			using (var unityPlayer = new AndroidJavaClass("com.unity3d.player.UnityPlayer"))
			{	using (var activity = unityPlayer.GetStatic<AndroidJavaObject>("currentActivity"))
				{	using (var window = activity.Call<AndroidJavaObject>("getWindow"))
					{	using (var view = window.Call<AndroidJavaObject>("getDecorView"))
						{	
							int SYSTEM_UI_FLAG_FULLSCREEN  = 4;
							int SYSTEM_UI_FLAG_HIDE_NAVIGATION = 2;
							int SYSTEM_UI_FLAG_IMMERSIVE_STICKY = 4096;
							view.Call ("setSystemUiVisibility", SYSTEM_UI_FLAG_FULLSCREEN |
							           SYSTEM_UI_FLAG_HIDE_NAVIGATION | SYSTEM_UI_FLAG_IMMERSIVE_STICKY);
						}
					}
				}
			}
		}
		#endregion
		
		#region Public Methods

		/// <summary> Proovides an easy way to instantiate gameObjects from the resource folder. </summary>
		/// <param name="path"> The path to the GameObject to instantiate. </param>
		/// <param name="name"> The new name of the resulting Gameobject instance. </param>
		/// <param name="parent"> The parent of the resulting Gameobject instance. </param>
		public static GameObject Instantiate (string path, string name, GameObject parent)
		{
			// Instantiate the GameObject
			GameObject gameObject;
			if (path == null || Resources.Load<GameObject>(path) == null) { gameObject = new GameObject(); }
			else { gameObject = (GameObject) Instantiate(Resources.Load<GameObject>(path)); }

			// Apply the optional parameters
			if (name != null) gameObject.name = name;
			if (parent != null)
			{
				Vector3 localPosition = gameObject.transform.localPosition;
				Quaternion localRotation = gameObject.transform.localRotation;
				Vector3 localScale = gameObject.transform.localScale;
				gameObject.transform.parent = parent.transform;
				gameObject.transform.localPosition = localPosition;
				gameObject.transform.localRotation = localRotation;
				gameObject.transform.localScale = localScale;
			}

			// Return the resulting GameObject instance
			return gameObject;
		}

		/// <summary> Loads a string from a resource file. </summary>
		/// <returns> The resulting string. </returns>
		/// <param name="path"> The file path of the resource. </param>
		public static string LoadString(string path)
		{
			// Get the referenced asset
			TextAsset asset = (TextAsset)Resources.Load(path);

			// make sure the asset exists
			if (asset == null) throw new IOException("File not found: " + path);

			// Return the asset text
			return asset.text;
		}

		/// <summary> Gets a Text resource associated to a widget </summary>
		/// <returns> The Text resource. </returns>
		/// <param name="widget"> The Widget node.</param>
		/// <param name="name"> The name of the resource. </param>
		public static Text GetResource(Widget widget, string name)
		{
			// The resouce locator
			string locator = "";

			Resource resource = null;

			if (widget != null)
			{

				// Look for the first node
				Widget searchWidget = widget;
				while (searchWidget != null)
				{
					if (!searchWidget.Resource.IsEmpty) 
					{
						if (locator.Length!=0) locator = "." + locator;
						locator = "resource." + searchWidget.Resource + locator;
					}

					// Continue with the parent node
					searchWidget = searchWidget.Ancestor<Widget>();
				}

				// Get the resource node
				resource = (Resource) widget.Ancestor<Platform>()[locator];
			}
			else
			{
				string[] locatorParts = name.Split(new string[]{"."}, StringSplitOptions.RemoveEmptyEntries);
				name = locatorParts[locatorParts.Length-1];
				resource = Root.Instance.Platforms[0].Resources[0].Resources[locatorParts[0]];
			}

			if (resource == null) resource = Root.Instance.Platforms[0].Resources[0];
			
			// Return the associated text node
			if (resource.Texts.Contains(name)) return resource.Texts[name];
			if (widget != null && resource.Texts.Contains(widget.Name + name)) return resource.Texts[widget.Name + name];
			if (widget != null && resource.Texts.Contains(widget.Name)) return resource.Texts[widget.Name];


			// If no appropiate resource was found return a null reference
			return null;
		}

		public delegate void DelayedFuntion();

		public static void Invoke(float time, DelayedFuntion delayedFunction)
		{
			Manager.Instance.StartCoroutine(Manager.Instance.InvokeAfterTime(time,delayedFunction));
		}

		private IEnumerator InvokeAfterTime(float waitTime, DelayedFuntion delayedFunction)
		{
			yield return new WaitForSeconds(waitTime);
			delayedFunction.DynamicInvoke();
		}

		public static void SaveResults()
		{
			// The common name of the files
			string fileName = "XUIMLExperiments" + DateTime.Now.ToString("yy.MM.dd-HH.mm.ss");

			// Show a message on console
			Debug.LogWarning("Saving results:" );

			// Calculate the SUS total score
			int susScore = MetricsManager.CalculateSUSScore();

			// Show a message on console
			Debug.LogWarning("SUS Score: " + susScore);

			// Save the camera texture
			if (CameraImage.FinalImage != null)
			{
				byte[] imageData = CameraImage.FinalImage.EncodeToJPG();
				FileStream imageFile = File.Open(Application.persistentDataPath + "/" + fileName + ".jpg", FileMode.Create);
				BinaryWriter imageWritter = new BinaryWriter(imageFile); imageWritter.Write(imageData); imageFile.Close();
			
				GetResource(null, "PhotoData").Update(Convert.ToBase64String(imageData));

				// Show a message on console
				Debug.LogWarning("Camera texture saved as " + fileName + ".jpg");
			} else Debug.LogWarning("Camera texture was null");

			// Show a message on console
			if (OutputManager.SoundExists("Salute"))
			{
				OutputManager.SaveSound("Salute", Application.persistentDataPath + "/" + fileName + ".wav");

				GetResource(null, "SaluteData").Update(Convert.ToBase64String(OutputManager.GetSoundData("Salute")));

				Debug.LogWarning("Microphone sound saved as " + fileName + ".wav");
			}
			else Debug.LogWarning("Microphone sound was null");

			// Update the resource node
			Resource resourceNode = Root.Instance.Platforms[0].Resources[0];
			resourceNode.Update();

			// FAILSAFE Serialize the entire resource
			Document failsafe = new Document(Application.persistentDataPath + "/" + fileName + ".xml", resourceNode);
			failsafe.Serialize();
			Debug.LogWarning("Failsafe Document saved as " + failsafe.FilePath);


			// Create a document and serialize it to upload its contents
			Document results = new Document("http://www.xuiml.org/experiments.php", resourceNode);
			results.Serialize(2);

			// Send the results to the server
			Debug.LogWarning("Results uploaded to the server");

			// Indicate that the results are saved
			GetResource(null, "ResultDataSaved").Update("True");
		}

		/// <summary> Finish the application execution. </summary>
		public static void QuitApplication() { Debug.LogWarning("QUIT"); Application.Quit();  }

		/// <summary> Determines if the execution is on Unity editor. </summary>
		/// <returns><c>true</c> if is editor; otherwise, <c>false</c>.</returns>
		public static bool OnUnityEditor()
		{
			#if UNITY_EDITOR
			return true;

			#else
			return false;
			#endif
		}

		#endregion
	}
}