using UnityEngine;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Defines a target for manipulation operations. </summary>
	public class ManipulationTarget : MonoBehaviour
	{
		#region Private Constants

		/// <summary> The translation threshold. </summary>
		private const float TranslationThreshold = 0.02f;

		/// <summary> The rotation threshold. </summary>
		private const float RotationThreshold = 60.0f;

		/// <summary> The scale threshold. </summary>
		private const float ScaleThreshold = 0.22f;

		#endregion

		#region Public Fields

		/// <summary> The widget node. </summary>
		Widget WidgetNode;

		/// <summary> The widget behaviour. </summary>
		WidgetBehaviour Widget;

		/// <summary> The widget GameObject. </summary>
		GameObject WidgetObject;

		/// <summary> The target node. </summary>
		Widget TargetWidgetNode;

		/// <summary> The target widget. </summary>
		WidgetBehaviour TargetWidget;

		/// <summary> The target widget GameObject. </summary>
		GameObject TargetWidgetObject;

		/// <summary> The target widget Manipulator. </summary>
		PivotPointManipulator TargetWidgetManipulator;

		WidgetBehaviour TargetWidgetPositioned;

		GameObject Guides;

		/// <summary> The size of the grid. </summary>
		private Vector3 GridSize = new Vector3(6f,2f,3f);

		#endregion
		
		#region Private Methods

		private void UpdateGuides(WidgetBehaviour widget)
		{ if (widget.Selected) Guides.animation.Play("ShowScale"); else Guides.animation.Play("HideScale"); }

		#endregion

		#region Public Methods
		
		public void Start()
		{
			// Get the widget properties
			Widget = GetComponent<WidgetBehaviour>();
			WidgetNode = Widget.WidgetNode;
			WidgetObject = this.gameObject;

			// Extract the manipulation target name
			string targetWidgetName = null;
			Behavior.Action action = WidgetNode.Behavior.Actions["initialize"];
			foreach (Behavior.Action.Reaction reaction in action.Reactions)
			{ if (reaction.Key == "ManipulationTarget") { targetWidgetName = reaction.Value; break; } }
			if (targetWidgetName == null) throw new System.Exception("No valid ManipulationTarget value found");

			// Get the target widget properties
			TargetWidgetNode = WidgetNode.Ancestor<Widget>().Widgets[targetWidgetName];
			TargetWidgetObject = ((Element)TargetWidgetNode.Representation).Existence;
			TargetWidget = TargetWidgetObject.GetComponent<WidgetBehaviour>();
			TargetWidgetManipulator = TargetWidget.GetComponentInChildren<PivotPointManipulator>();

			foreach (Transform child in this.transform) if (child.name == "Guides") Guides = child.gameObject;

			Vector3 size = this.renderer.bounds.size * 10f;
			foreach (Transform child in Guides.transform)
			{
				child.localScale = size;

				switch (child.name)
				{
				case "GuideX": 
					child.localScale = size + new Vector3(GridSize.x, 0.0f, 0.0f);
					child.localPosition = new Vector3(-this.transform.position.x, 0.0f, 0.0f);
					break;
				case "GuideY": 
					child.localScale = Vector3.zero; // 
//					child.localScale = size + new Vector3(0.0f, GridSize.y, 0.0f);
//					child.localPosition = new Vector3(0, -this.transform.position.y, 0.0f);
					break;
				case "GuideZ":
					child.localScale = size + new Vector3(0.0f, 0.0f, GridSize.z);
					child.localPosition = new Vector3(0.0f, 0.0f, -this.transform.position.z);
					break;
				}
			}

			TargetWidget.SelectedStateChanged += UpdateGuides;
		}


		public void Update()
		{
			// If the manipulation widget or the target widget not enabled, do nothing
			if (!Widget.Enabled || !TargetWidget.Enabled) return;

			if (TargetWidget.Selected)
			{
				Transform currentTransform = WidgetObject.transform, targetTransform = TargetWidgetObject.transform;

				if (Vector3.Distance(currentTransform.position, targetTransform.position) < TranslationThreshold &&
				    Quaternion.Angle(currentTransform.rotation, targetTransform.rotation) < RotationThreshold &&
				    Vector3.Distance(currentTransform.lossyScale, targetTransform.lossyScale) < ScaleThreshold)
				{
					Widget.React("activate"); 
					TargetWidget.Selected = false;
					TargetWidget.Activation = WidgetBehaviour.ActivationMode.None;
					TargetWidgetManipulator.Deactivate();


					TargetWidgetObject.transform.position = WidgetObject.transform.position;
					TargetWidgetObject.transform.rotation = WidgetObject.transform.rotation;
					TargetWidgetObject.transform.localScale = WidgetObject.transform.localScale;

					// FIX
					TargetWidgetNode.Size.Update(0.1f, 0.1f, 0.1f);
					TargetWidgetNode.Size.Update(true);
					TargetWidgetNode.Position.Update(Element.ToVector(WidgetObject.transform.position));
					TargetWidgetNode.Rotation.Update(Element.ToVector(WidgetObject.transform.eulerAngles));
					TargetWidgetNode.Position.Update(true);


					// Apply the new material
					this.renderer.material = Resources.Load<UnityEngine.Material>
						("Themes/Futuristic/Materials/ManipulationAchieved");

					// Update the material
					Unity.Element.UpdateMaterials(this.gameObject);

					// Play a Sound
					OutputManager.PlaySound("ManipulationAchieved");
				}
			}
		}

		#endregion
	}
}