using UnityEngine;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Pivot-Point-based manipulator. </summary>
	public class PivotPointManipulator : Manipulator
	{
		#region Private Fields

		/// <summary> The current manipulation operation. </summary>
		private ManipulationOperation CurrentManipulationOperation;

		/// <summary> The current manipulation operation node. </summary>
		private Text CurrentManipulationOperationNode;

		/// <summary> The translation manipulator GameObject. </summary>
		private GameObject TranslateManipulator;

		/// <summary> The rotation manipulator GameObject. </summary>
		private GameObject RotateManipulator;

		/// <summary> The scale manipulator GameObject. </summary>
		private GameObject ScaleManipulator;

		/// <summary> The bounding box GameObject. </summary>
		private GameObject BoundingBox;

		/// <summary> The size of the grid. </summary>
		private Vector3 GridSize = new Vector3(0.60f, 0.30f, 0.30f);

		/// <summary> The grid minimum position. </summary>
		private Vector3 GridMinPosition;

		/// <summary> The grid max position. </summary>
		private Vector3 GridMaxPosition;

		/// <summary> The size of any grid cell. </summary>
		private float GridCellSize = 0.01f;

		/// <summary> The manipulator helper. </summary>
		private GameObject ManipulatorHelper;

		/// <summary> The snapping angle of the rotation manipulator. </summary>
		private float RotateManipulatorSnapAngle = 15.0f;

		/// <summary> The initial rotation manipulator angle. </summary>
		private float RotateManipulatorInitialAngle = float.NaN;

		/// <summary> The size of the rotate manipulator colliders. </summary>
		private float RotateManipulatorSize = 0.1f;

		/// <summary> The width of the rotate manipulator colliders. </summary>
		private float RotateManipulatorWidth = 0.04f;

		/// <summary> The initial rotation manipulator value. </summary>
		private Vector3 RotateManipulatorInitialAngles = new Vector3(0, 0, 0);

		/// <summary> The minimum size of the scale manipulator colliders. </summary>
		private Vector3 ScaleManipulatorMinSize = new Vector3(0.05f, 0.05f, 0.05f);

		/// <summary> The maximum size of the scale manipulator colliders. </summary>
		private Vector3 ScaleManipulatorMaxSize = new Vector3(0.1f, 0.1f, 0.1f);

		/// <summary> The minimum size of the manipulator. </summary>
		private float ManipulatorMinSize = 0.0001f;
		
		/// <summary> The maximum size of the manipulator. </summary>
		private float ManipulatorMaxSize = 1.0f;

		/// <summary> The colliders associated to the component. </summary>
		private List<Collider> ComponentColliders;

		#endregion
		
		#region Private Methods
		
		/// <summary> Initializes the instance. </summary>
		private void Start()
		{
			// Initialize the manipulator
			Initialize();
			
			// Get the different manipulators
			foreach (Transform child in this.transform)
			{
				if (child.name == "Translate") TranslateManipulator = child.gameObject;
				if (child.name == "Rotate") RotateManipulator = child.gameObject;
				if (child.name == "Scale") ScaleManipulator = child.gameObject;
				if (child.name == "BoundingBox") BoundingBox = child.gameObject;
			}
			
			// Hide all manipulators
			TranslateManipulator.animation.Play("HideScaleFast");
			RotateManipulator.animation.Play("HideScaleFast");
			ScaleManipulator.animation.Play("HideScaleFast");

			// Set the initial transform operation
			CurrentManipulationOperation = ManipulationOperation.None;
			CurrentManipulationOperationNode = Root.Instance.Platforms[0].Resources[0].Texts["CurrentOperation"];
			CurrentManipulationOperationNode.UpdatePerformed += this.SetManipulationOperation;
			SetManipulationOperation(CurrentManipulationOperationNode, true);

			ComponentColliders = new List<Collider>(GetComponentsInChildren<Collider>());

			// Calculate the minimum and maximum positions in the grid
			GridMinPosition = new Vector3(-GridSize.x / 2f, 0, -GridSize.z / 2f);
			GridMaxPosition = new Vector3(GridSize.x / 2f, GridSize.y, GridSize.z / 2f);
		}


		/// <summary> Called after rendering. </summary>
		private void LateUpdate()
		{
			// Check if the AR camera is ready
			if (AugmentedReality.CameraObject == null) return;

			// FIX Make absolutely sure that we are in the right section
			if (Root.Instance.Platforms[0].Resources[0].Texts["Section"] != "UseCase2Execution")
			{ transform.parent.localScale = Vector3.zero; this.transform.localScale = Vector3.zero; return; }

			//
			if ((!Widget.Selected && !Widget.Enabled) && !animation.isPlaying)
			{ this.transform.localScale = Vector3.zero; return; }

			// Keep the transform manipulator always aligned
			TranslateManipulator.transform.eulerAngles = Vector3.zero;

			//Calculate the current scale factor
			/// <summary> The current scale factor (required to sustain interaction even at long distances. </summary>
			float scaleFactor = Vector3.Distance(transform.position, AugmentedReality.CameraObject.transform.position);

			// 
			Vector3 parentScale = WidgetObject.transform.localScale;
			Vector3 widgetScale = new Vector3(WidgetNode.Size.X, WidgetNode.Size.Y, WidgetNode.Size.Z) * 10.0f;

			if (WidgetObject.animation.isPlaying) WidgetObject.transform.localScale = Cross(parentScale, widgetScale);
			else WidgetObject.transform.localScale = widgetScale;

			//
			if (CurrentManipulationOperation == ManipulationOperation.Rotate && ManipulatorHelper != null)
				ManipulatorHelper.transform.localScale = Vector3.one * scaleFactor;

			if (CurrentManipulationOperation == ManipulationOperation.Scale && !ScaleManipulator.animation.isPlaying)
				ScaleManipulator.transform.localScale = parentScale;

			if (!animation.isPlaying) this.transform.localScale = NormalizedScale(gameObject) * scaleFactor;
			else this.transform.localScale = NormalizedScale(gameObject) * this.transform.localScale.x * scaleFactor;

			// Only selected widgets can be seen
			if (!Widget.Selected && !animation.isPlaying) { this.transform.localScale = Vector3.zero; return; }

			BoundingBox.transform.position = WidgetObject.renderer.bounds.center;
			BoundingBox.transform.rotation = this.transform.rotation;
			BoundingBox.transform.localScale = Cross(NormalizedScale(BoundingBox) * 10.0f, WidgetObject.renderer.bounds.size );
			BoundingBox.transform.eulerAngles = Vector3.zero;

			// Update the time metrics
			float deltaTime = Time.deltaTime;
			switch (CurrentManipulationOperation)
			{
			case ManipulationOperation.Translate: MetricsManager.AddTime(MetricsManager.TaskId.UC2A, deltaTime); break;
			case ManipulationOperation.Rotate: MetricsManager.AddTime(MetricsManager.TaskId.UC2B, deltaTime); break;
			case ManipulationOperation.Scale: MetricsManager.AddTime(MetricsManager.TaskId.UC2C, deltaTime); break;
			}
		}

		private Vector3 Cross (Vector3 v1, Vector3 v2) { return new Vector3(v1.x * v2.x, v1.y * v2.y, v1.z * v2.z); }

		private Vector3 Invert (Vector3 v) { return new Vector3(1.0f / v.x, 1.0f / v.y, 1.0f / v.z); }

		private Vector3 NormalizedScale(GameObject go)
		{ if (go.transform.parent.lossyScale == Vector3.zero) return Vector3.zero; return Invert(go.transform.parent.lossyScale); }

		private void SetManipulationOperation(Node node = null) { SetManipulationOperation(node, false); }

		private void SetManipulationOperation(Node node, bool forced)
		{
			ManipulationOperation newManipulationOperation = ManipulationOperation.None;
			if (CurrentManipulationOperationNode.Value== "Translate") newManipulationOperation = ManipulationOperation.Translate;
			else if (CurrentManipulationOperationNode.Value == "Rotate") newManipulationOperation = ManipulationOperation.Rotate;
			else if (CurrentManipulationOperationNode.Value == "Scale") newManipulationOperation = ManipulationOperation.Scale;
			

			if (CurrentManipulationOperation == newManipulationOperation && !forced) return;
			
			switch (CurrentManipulationOperation)
			{
			case ManipulationOperation.Translate: TranslateManipulator.animation.Play("HideScaleFast"); break;
			case ManipulationOperation.Rotate: RotateManipulator.animation.Play("HideScaleFast"); break;
			case ManipulationOperation.Scale: ScaleManipulator.animation.Play("HideScaleFast"); break;
			}

			switch (newManipulationOperation)
			{
			case ManipulationOperation.Translate: TranslateManipulator.animation.Play("ShowScaleFast"); break;
			case ManipulationOperation.Rotate: RotateManipulator.animation.Play("ShowScaleFast"); break;
			case ManipulationOperation.Scale: ScaleManipulator.animation.Play("ShowScaleFast"); break;
			}

			// Update the transfom operation
			CurrentManipulationOperation = newManipulationOperation;
		}


		/// <summary> Creates the translation helper. </summary>
		/// <param name="component"> The manipulator component. </param>
		private void CreateTranslationHelper(ManipulatorComponent component)
		{
			Vector3 widgetPosition = WidgetObject.transform.position;
			Vector3 hitPosition = InputManager.Hit.point;
			Vector3 positionDisplacement = (hitPosition - widgetPosition);

			ManipulatorHelper = new GameObject("ManipulatorHelper");
			ManipulatorHelper.layer = WidgetObject.layer;
			ManipulatorHelper.transform.position = positionDisplacement;

			float minValue = 0, maxValue= 0;
			switch (component.Axes)
			{
			case Axes.X:
				minValue = (GridMinPosition.x - widgetPosition.x); 
				maxValue = (GridMaxPosition.x - widgetPosition.x);
				break;
			case Axes.Y:
				minValue = (GridMinPosition.y - widgetPosition.y); 
				maxValue = (GridMaxPosition.y - widgetPosition.y);
				break;
			case Axes.Z:
				minValue = (GridMinPosition.z - widgetPosition.z); 
				maxValue = (GridMaxPosition.z - widgetPosition.z);
				break;
			}

			for (float value = minValue; value <= maxValue; value += GridCellSize)
			{
				GameObject translatorHelper = new GameObject("TranslatorHelper");
				translatorHelper.transform.parent = ManipulatorHelper.transform;
				translatorHelper.layer = WidgetObject.layer;
				ManipulatorComponent manipulatorComponent = translatorHelper.AddComponent<ManipulatorComponent>();
				manipulatorComponent.ParentManipulator = this;
				BoxCollider collider1 = translatorHelper.AddComponent<BoxCollider>();
				BoxCollider collider2 = translatorHelper.AddComponent<BoxCollider>();

				switch (component.Axes)
				{
				case Axes.X:
					translatorHelper.transform.localPosition = new Vector3(value, 0 ,0) + widgetPosition;
					collider1.size = new Vector3(GridCellSize, ManipulatorMinSize, ManipulatorMaxSize);
					collider2.size = new Vector3(GridCellSize, ManipulatorMaxSize, ManipulatorMinSize);
					break;
				case Axes.Y:
					translatorHelper.transform.localPosition = new Vector3(0, value ,0) + widgetPosition;
					collider1.size = new Vector3(ManipulatorMinSize, GridCellSize, ManipulatorMaxSize);
					collider2.size = new Vector3(ManipulatorMaxSize, GridCellSize, ManipulatorMinSize);
					break;
				case Axes.Z:
					translatorHelper.transform.localPosition = new Vector3(0, 0 ,value) + widgetPosition;
					collider1.size = new Vector3(ManipulatorMinSize, ManipulatorMaxSize, GridCellSize);
					collider2.size = new Vector3(ManipulatorMaxSize, ManipulatorMinSize, GridCellSize);
					break;
				}

				DebugConsole.DrawBoxCollider(collider1, 10); DebugConsole.DrawBoxCollider(collider2, 10);
			}
		}

		/// <summary> Creates the rotation helper. </summary>
		/// <param name="component"> The manipulator component. </param>
		private void CreateRotationHelper(ManipulatorComponent component)
		{
			Vector3 widgetPosition = WidgetObject.transform.position;

			ManipulatorHelper = new GameObject("ManipulatorHelper");
			ManipulatorHelper.layer = WidgetObject.layer;

			float bladeLength = Mathf.Cos(Mathf.Deg2Rad * RotateManipulatorSnapAngle) * RotateManipulatorSize;
			float bladeWidth = Mathf.Sin(Mathf.Deg2Rad * RotateManipulatorSnapAngle) * RotateManipulatorSize / 2f;
			float bladeDepth = RotateManipulatorWidth / 2f;

			UnityEngine.Mesh bladeMesh = new UnityEngine.Mesh();
			bladeMesh.name = "BladeMesh";

			Axes rotationAxis = component.Axes;

			switch (rotationAxis)
			{
			case Axes.X:
				bladeMesh.vertices = new Vector3[7]
				{
					new Vector3(0f, 0f, 0f), 
					new Vector3(0f, bladeLength, -bladeWidth) * 100f,
					new Vector3(0f, bladeLength, bladeWidth) * 100f,
					new Vector3(-bladeDepth, bladeLength, -bladeWidth),
					new Vector3(-bladeDepth, bladeLength, +bladeWidth),
					new Vector3(+bladeDepth, bladeLength, -bladeWidth),
					new Vector3(+bladeDepth, bladeLength, +bladeWidth)
				};
				break;
			case Axes.Y:
				bladeMesh.vertices = new Vector3[7]
				{
					new Vector3(0f, 0f, 0f), 
					new Vector3(bladeLength, 0f, -bladeWidth) * 100f,
					new Vector3(bladeLength, 0f, bladeWidth) * 100f,
					new Vector3(bladeLength, -bladeDepth, -bladeWidth),
					new Vector3(bladeLength, -bladeDepth, +bladeWidth),
					new Vector3(bladeLength, +bladeDepth, -bladeWidth),
					new Vector3(bladeLength, +bladeDepth, +bladeWidth)
				};
				break;
			case Axes.Z:
				bladeMesh.vertices = new Vector3[7]
				{
					new Vector3(0f, 0f, 0f), 
					new Vector3(bladeLength, -bladeWidth, 0f) * 100f,
					new Vector3(bladeLength, bladeWidth, 0f) * 100f,
					new Vector3(bladeLength, -bladeWidth, -bladeDepth),
					new Vector3(bladeLength, +bladeWidth, -bladeDepth),
					new Vector3(bladeLength, -bladeWidth, +bladeDepth),
					new Vector3(bladeLength, +bladeWidth, +bladeDepth)
				};
				break;
			}

			bladeMesh.triangles = new int[18] {
				0, 1, 2, 0, 2, 1,
				0, 4, 3, 0, 5, 6,
				3, 4, 5, 4, 6, 5
			};
			bladeMesh.RecalculateNormals();

			for (float angle = RotateManipulatorSnapAngle; angle <= 360; angle += RotateManipulatorSnapAngle)
			{
				GameObject rotationHelper = new GameObject("RotationHelper " + angle.ToString());
				rotationHelper.transform.parent = ManipulatorHelper.transform;
				rotationHelper.layer = WidgetObject.layer;
				ManipulatorComponent manipulatorComponent = rotationHelper.AddComponent<ManipulatorComponent>();
				manipulatorComponent.ParentManipulator = this;
				manipulatorComponent.Axes = component.Axes; // Critical
				MeshCollider bladeCollider = rotationHelper.AddComponent<MeshCollider>();
				bladeCollider.sharedMesh = bladeMesh;

				// Debug
//				MeshFilter bladeFilter = rotationHelper.AddComponent<MeshFilter>();
//				bladeFilter.sharedMesh = bladeMesh;
//				MeshRenderer bladeRenderer = rotationHelper.AddComponent<MeshRenderer>();
//				bladeRenderer.material.color = new UnityEngine.Color(angle/360.0f, 0.5f, 0.5f);

				switch (rotationAxis)
				{
				case Axes.X: rotationHelper.transform.localEulerAngles = new Vector3(angle, 0, 0); break;
				case Axes.Y: rotationHelper.transform.localEulerAngles = new Vector3(0, angle, 0); break;
				case Axes.Z: rotationHelper.transform.localEulerAngles = new Vector3(0, 0, angle); break;
				}

			}
			ManipulatorHelper.transform.position = widgetPosition;
			ManipulatorHelper.transform.rotation = WidgetObject.transform.rotation;
		}

		/// <summary> Creates the scale helper. </summary>
		/// <param name="component"> The manipulator component. </param>
		private void CreateScaleHelper(ManipulatorComponent component)
		{
			Vector3 widgetPosition = WidgetObject.transform.position;
			Vector3 hitPosition = InputManager.Hit.point;
			Vector3 positionDisplacement = (hitPosition - widgetPosition);
			Vector3 widgetSize = new Vector3(WidgetNode.Size.X, WidgetNode.Size.Y, WidgetNode.Size.Z);

			ManipulatorHelper = new GameObject("ManipulatorHelper");
			ManipulatorHelper.layer = WidgetObject.layer;
			ManipulatorHelper.transform.position = positionDisplacement;

			float minValue = 0, maxValue= 0;
			switch (component.Axes)
			{
			case Axes.X: minValue = ScaleManipulatorMinSize.x; maxValue = ScaleManipulatorMaxSize.x; break;
			case Axes.Y: minValue = ScaleManipulatorMinSize.y; maxValue = ScaleManipulatorMaxSize.y; break;
			case Axes.Z: minValue = ScaleManipulatorMinSize.z; maxValue = ScaleManipulatorMaxSize.z; break;
			}


			for (float value = minValue; value <= maxValue; value += GridCellSize)
			{
				GameObject scaleHelper = new GameObject("ScaleHelper " + value);
				scaleHelper.transform.parent = ManipulatorHelper.transform;
				scaleHelper.layer = WidgetObject.layer;
				ManipulatorComponent manipulatorComponent = scaleHelper.AddComponent<ManipulatorComponent>();
				manipulatorComponent.ParentManipulator = this;
				manipulatorComponent.Axes = component.Axes; // Critical
				BoxCollider collider1 = scaleHelper.AddComponent<BoxCollider>();
				BoxCollider collider2 = scaleHelper.AddComponent<BoxCollider>();

				switch (component.name)
				{
				case "ScaleXN":
					scaleHelper.transform.localPosition = new Vector3(value - widgetSize.x, 0 ,0) + widgetPosition;
					collider1.size = new Vector3(GridCellSize, ManipulatorMinSize, ManipulatorMaxSize);
					collider2.size = new Vector3(GridCellSize, ManipulatorMaxSize, ManipulatorMinSize);
					break;
				case "ScaleXP": // Accout for the 
					scaleHelper.transform.localPosition = new Vector3(-(value - widgetSize.x), 0 ,0) + widgetPosition;
					collider1.size = new Vector3(GridCellSize, ManipulatorMinSize, ManipulatorMaxSize);
					collider2.size = new Vector3(GridCellSize, ManipulatorMaxSize, ManipulatorMinSize);
					break;
				case "ScaleYP":
					scaleHelper.transform.localPosition = new Vector3(0, value - widgetSize.y ,0) + widgetPosition;
					collider1.size = new Vector3(ManipulatorMinSize, GridCellSize, ManipulatorMaxSize);
					collider2.size = new Vector3(ManipulatorMaxSize, GridCellSize, ManipulatorMinSize);
					break;
				case "ScaleYN":
					scaleHelper.transform.localPosition = new Vector3(0, -(value - widgetSize.y),0) + widgetPosition;
					collider1.size = new Vector3(ManipulatorMinSize, GridCellSize, ManipulatorMaxSize);
					collider2.size = new Vector3(ManipulatorMaxSize, GridCellSize, ManipulatorMinSize);
					break;
				case "ScaleZN":
					scaleHelper.transform.localPosition = new Vector3(0, 0 ,-(value - widgetSize.z)) + widgetPosition;
					collider1.size = new Vector3(ManipulatorMinSize, ManipulatorMaxSize, GridCellSize);
					collider2.size = new Vector3(ManipulatorMaxSize, ManipulatorMinSize, GridCellSize);
					break;
				case "ScaleZP":
					scaleHelper.transform.localPosition = new Vector3(0, 0 ,value - widgetSize.z) + widgetPosition;
					collider1.size = new Vector3(ManipulatorMinSize, ManipulatorMaxSize, GridCellSize);
					collider2.size = new Vector3(ManipulatorMaxSize, ManipulatorMinSize, GridCellSize);
					break;
				}
//				DebugConsole.DrawBoxCollider(collider1, 10); DebugConsole.DrawBoxCollider(collider2, 10);
			}
		}

		#endregion
		
		#region Public Methods

		/// <summary> Activates the manipulator. </summary>
		public override void Activate()	{ SetManipulationOperation(CurrentManipulationOperationNode); }

		/// <summary> Deactivates the manipulator. </summary>
		public override void Deactivate()
		{ 
			// Destroy the manipulator helper
			if (ManipulatorHelper != null) Destroy(ManipulatorHelper);
			
			// Disable the colliders
			WidgetObject.collider.enabled = false;
			foreach (Collider componentCollider in ComponentColliders) componentCollider.enabled = false;

		}

		/// <summary> Updates the manipulation operation. </summary>
		/// <param name="component">The associated manipulator component.</param>
		public override void UpdateManipulationOperation(ManipulatorComponent component, RaycastHit hit)
		{
//			DebugConsole.DrawStar(InputManager.HitPosition, 0.1f,1,"green");
//			Debug.LogWarning("Action: " + InputManager.CurrentAction);

			// Get the list of "sibling" manipulator components associated to the manipulator
			ManipulatorComponent[] siblingComponents = component.transform.parent
				.GetComponentsInChildren<ManipulatorComponent>();

			switch (InputManager.CurrentAction)
			{
			case InputManager.Action.Press:
				
				// Disable the component colliders
				WidgetObject.collider.enabled = false;
				foreach (Collider componentCollider in ComponentColliders) componentCollider.enabled = false;

				// Make sure the manipulator helper have been destroyed
				if (ManipulatorHelper != null) Destroy(ManipulatorHelper);

				// Create the manipulation helpers
				switch (CurrentManipulationOperation)
				{
				case ManipulationOperation.Translate: CreateTranslationHelper(component); break;
				case ManipulationOperation.Rotate: CreateRotationHelper(component); break;
				case ManipulationOperation.Scale: CreateScaleHelper(component); break;
				}

				if (CurrentManipulationOperation == ManipulationOperation.Rotate)
				{
					RotateManipulatorInitialAngle = float.NaN;
					RotateManipulatorInitialAngles = new Vector3(WidgetNode.Rotation.X, WidgetNode.Rotation.Y, WidgetNode.Rotation.Z);
				}

				// Hide the sibling components that do not share the same axes
				foreach (ManipulatorComponent siblingComponent in siblingComponents)
					if (siblingComponent.Axes != component.Axes) siblingComponent.transform.localScale = Vector3.zero;

				// Check for errors
				switch (CurrentManipulationOperation)
				{
				case ManipulationOperation.Rotate: 
					if (WidgetObject.name != "ManipulationObject3") MetricsManager.AddError(MetricsManager.TaskId.UC2B);
					break;
				case ManipulationOperation.Scale:
					if (WidgetObject.name != "ManipulationObject1") MetricsManager.AddError(MetricsManager.TaskId.UC2C);
					break;
				}

				break;
			case InputManager.Action.Move:

				// Make sure the current collider belongs to the manipulator helper
				if (ManipulatorHelper == null || hit.collider == null || hit.collider.transform.parent == null ||
				    hit.collider.transform.parent != ManipulatorHelper.transform) return;

				switch (CurrentManipulationOperation)
				{
				case ManipulationOperation.Translate:

					// Get the current collider
					BoxCollider translationCollider = (BoxCollider) hit.collider;

					// Get the current position
					Vector3 position = translationCollider.transform.localPosition;

					WidgetNode.Position.X.Update(position.x);
					WidgetNode.Position.Y.Update(position.y);
					WidgetNode.Position.Z.Update(position.z);

					// Debug.LogWarning("Position: " + position.ToString());

					// Update the node
					WidgetNode.Update();

					break;
				case ManipulationOperation.Rotate:

					// Get the current collider
					MeshCollider rotationCollider = (MeshCollider) hit.collider;

					// Parse the rotation value
					float rotation = float.Parse(rotationCollider.name.Split(new char[]{' '})[1]);

					if (float.IsNaN(RotateManipulatorInitialAngle)) RotateManipulatorInitialAngle = rotation;

					rotation -= RotateManipulatorInitialAngle;

					// Apply the value to the correct scale
					switch (component.Axes)
					{
					case Axes.X : WidgetNode.Rotation.X.Update(RotateManipulatorInitialAngles.x + rotation); break;
					case Axes.Y : WidgetNode.Rotation.Y.Update(RotateManipulatorInitialAngles.y + rotation); break;
					case Axes.Z : WidgetNode.Rotation.Z.Update(RotateManipulatorInitialAngles.z + rotation); break;
					}

//					Debug.LogWarning("Rotation: " + component.Axes + " " + rotation.ToString());

					// Update the node
					WidgetNode.Update();

					break;
				case ManipulationOperation.Scale:

					// Get the current collider
					BoxCollider scaleCollider = (BoxCollider) hit.collider;

					// Parse the scale value
					float scale = float.Parse(scaleCollider.name.Split(new char[]{' '})[1]);

					// Apply the value to the correct scale
					switch (component.Axes)
					{
					case Axes.X : WidgetNode.Size.X.Update(scale); break;
					case Axes.Y : WidgetNode.Size.Y.Update(scale); break;
					case Axes.Z : WidgetNode.Size.Z.Update(scale); break;
					}

					// Debug.LogWarning("Scale: " + component.Axes + " " + scale.ToString());

					// Update the node
					WidgetNode.Update();

					break;
				}

				break;
			case InputManager.Action.Release:
				
				// Destroy the manipulator helper
				Destroy(ManipulatorHelper);

				// Re-enable the colliders
				WidgetObject.collider.enabled = true;
				foreach (Collider componentCollider in ComponentColliders) componentCollider.enabled = true;

				// Show the sibling components
				foreach (ManipulatorComponent siblingComponent in siblingComponents)
					if (siblingComponent.Axes != component.Axes) siblingComponent.transform.localScale = Vector3.one;

				break;
			}
			
		}
		
		
		#endregion
		
	}
}