using UnityEngine;
using System;
using System.Collections.Generic;
using XUIML.Data;

namespace XUIML.Unity
{
	///<summary> Handles the general metrics of the experiment. </summary>
	public class MetricsManager : MonoBehaviour
	{
		#region Private Constants

		/// <summary> The task letters. </summary>
		private const string TaskLetters = "ABC";

		/// <summary> The minimum number of seconds that the marker has to be "lost" before starting 
		/// to count as an error.</summary>
		private const float MarkerLostErrorTime = 5.0f;

		#endregion

		#region Public Enumerations

		/// <summary> Enumerates the task IDs. </summary>
		public enum TaskId {UC0A, UC0B, UC0C, UC1A, UC1B, UC1C, UC2A, UC2B, UC2C, UC3A, UC3B, UC3C};

		#endregion

		#region Private Subclasses

		/// <summary> The data associated to each Use Case. </summary>
		private class UseCaseData
		{
			/// <summary> The time value. </summary>
			public float TimeValue; 

			/// <summary> The time counter node. </summary>
			public Text TimeNode; 

			/// <summary> Initializes a new instance of the UseCaseData class. </summary>
			/// <param name="time"> The time counter node. </param>
			public UseCaseData(Text time) { TimeValue = 0; TimeNode = time;}
		}

		/// <summary> The data associated to each task. </summary>
		private class TaskData
		{
			/// <summary> The time value. </summary>
			public float TimeValue;

			/// <summary> The time counter node.. </summary>
			public Text TimeNode;

			/// <summary> The error counter node.. </summary>
			public Text ErrorsNode;
			
			/// <summary> Initializes a new instance of the UseCaseData class. </summary>
			/// <param name="time">The time counter node. </param>
			/// <param name="time">The error counter node. </param>
			public TaskData(Text time, Text errors) { TimeValue = 0; TimeNode = time; ErrorsNode = errors;}
		}
		#endregion

		#region Private Fields

		/// <summary> The node that defines the current application section. </summary>
		private static Text CurrentSection;

		/// <summary> The collection of use cases. </summary>
		private static List<UseCaseData> UseCases;

		/// <summary> The collection of tasks. </summary>
		private static Dictionary<TaskId, TaskData> Tasks;

		/// <summary> The node that indicates if the marker is required. </summary>
		private Text MarkerRequired;

		/// <summary> The node that indicates if the marker is detected. </summary>
		private Text MarkerDetected;

		/// <summary> The number of seconds that the marker is currently being "lost". </summary>
		private float MarkerLostTime;

		#endregion

		#region Private Methods

		/// <summary> Initializes this instance. </summary>
		private void Start()
		{
			// Get the node with the current section name
			CurrentSection = Manager.GetResource (null, "Section");

			// Initialize the data collections
			UseCases = new List<UseCaseData> ();
			Tasks = new Dictionary<TaskId, TaskData> ();

			// Create the use case data
			for (int i = 0; i < 4; i++) UseCases.Add(new UseCaseData(Manager.GetResource(null,"UseCase" + i + "Time")));

			// Create the task data
			Array taskIds = Enum.GetValues(typeof(TaskId));
			for (int i = 0; i < 12; i++)
			{
				int useCaseNumber = i / 3; char taskLetter = TaskLetters[i % 3];
				Text timeNode = Manager.GetResource(null, "UseCase" + useCaseNumber + "" + taskLetter + "Time");
				Text errorsNode = Manager.GetResource(null, "UseCase" + useCaseNumber + "" + taskLetter + "Errors");
				Tasks.Add((TaskId)taskIds.GetValue(i), new TaskData(timeNode, errorsNode));
			}

			// Define the variables that manage physical navigation
			MarkerLostTime = 0;
			MarkerRequired = Manager.GetResource (null, "MarkerRequired");
			MarkerDetected = Manager.GetResource (null, "MarkerDetected");
		}

		/// <summary> Called on frame rendering. </summary>
		public void Update()
		{
			// get the current delta time
			float deltaTime = Time.deltaTime;

			string currentSectionName = CurrentSection.Value;

			// Check the current section value to update their associated time counters
			switch (currentSectionName)
			{
			case "EULA": case "GeneralInstructions": case "UseCase1Instructions" : case "UseCase2Instructions" :
			case "UseCase3Instructions": case "ThanksSection":
				AddTime(0, deltaTime); AddTime(TaskId.UC0A, deltaTime); break;
			case "UseCase1UserMetrics": case "UseCase2UserMetrics": case "UseCase3UserMetrics":
			case "UseCase0UserMetrics":
				AddTime(0, deltaTime); AddTime(TaskId.UC0B, deltaTime); break;
			case "UseCase1Execution": AddTime(0, deltaTime); AddTime(1, deltaTime); break;
			case "UseCase2Execution": AddTime(0, deltaTime); AddTime(2, deltaTime); break;
			case "UseCase3Execution": AddTime(0, deltaTime); AddTime(3, deltaTime); break;
			}

			// Check counter
			if (MarkerRequired.Value == "True")
			{
				// Add the time to the task
				AddTime(TaskId.UC0C, deltaTime);

				// Check if the marker is being detected
				if (MarkerDetected.Value == "True") MarkerLostTime = 0;
				else 
				{
					// Increase the time counter
					MarkerLostTime += deltaTime;

					// If the time counter exceeds a given number of seconds, mark it as an error
					if (MarkerLostTime > MarkerLostErrorTime) { AddError(TaskId.UC0C); MarkerLostTime = 0; }
				}
			}
		}

		#endregion

		#region Public methods

		/// <summary> Adds an ammount of time to an specific use case. </summary>
		/// <param name="uc"> The use case identifier.</param>
		/// <param name="time"> The ammount of time  to add. </param>
		public static void AddTime(int uc, float time)
		{ UseCases[uc].TimeValue += time; UseCases[uc].TimeNode.Update(((int)(UseCases[uc].TimeValue)).ToString()); }


		/// <summary> Adds an ammount of time to an specific task. </summary>
		/// <param name="taskId">The task identifier.</param>
		/// <param name="time"> The ammount of time  to add. </param>
		public static void AddTime(TaskId taskId, float time)
		{
			// FIX: Make sure that the task times are only inputed during the right Use Case
			switch(taskId)
			{
			case TaskId.UC1A: case TaskId.UC1B: case TaskId.UC1C: if (CurrentSection.Value!="UseCase1Execution") return; break;
			case TaskId.UC2A: case TaskId.UC2B: case TaskId.UC2C: if (CurrentSection.Value!="UseCase2Execution") return; break;
			case TaskId.UC3A: case TaskId.UC3B: case TaskId.UC3C: if (CurrentSection.Value!="UseCase3Execution") return; break;
			}

			// Add the time value
			Tasks[taskId].TimeValue += time; Tasks[taskId].TimeNode.Update(((int)(Tasks[taskId].TimeValue)).ToString());
		}


		/// <summary> Adds an error value to an specific task. </summary>
		/// <param name="taskId">The task identifier.</param>
		public static void AddError(TaskId taskId)
		{ Tasks[taskId].ErrorsNode.Update((int.Parse(Tasks[taskId].ErrorsNode.Value) + 1).ToString()); }


		/// <summary> Calculates the SUS total score. </summary>
		public static int CalculateSUSScore()
		{
			// Create a variable to store the total score
			int totalScore = 0;

			// Extract the sus values from the nodes
			for (int index = 0; index < 10; index++)
			{
				// Create the node name
				string nodeName = string.Format("SystemUsabilityScale.SUSPoint{0:00}Value", (index + 1));

				// Load the node value
				int value = int.Parse(Manager.GetResource(null, nodeName)) - 1;

				// Even values must be inverted
				if (index % 2 > 0) value = 4 - value;

				// Add the value to the total score
				totalScore += value;
			}

			// Multiply the score by 2.5 to adjust it to a [0..100] scale
			totalScore = (int)(totalScore * 2.5f);

			// Save the score in the appropriate node
			Manager.GetResource(null, "SystemUsabilityScale.SUSTotalScore").Update(totalScore.ToString());

			// Return the resulting score
			return totalScore;
		}

		#endregion
	}
}