using UnityEngine;
using System;
using System.IO;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	///<summary> Handles output signals. </summary>
	public class OutputManager : MonoBehaviour
	{

		private static string SoundsResourcePath = "Themes/Futuristic/Sounds/";

		private static GameObject Speaker;

		private static Dictionary <string, AudioClip> Sounds;


		public static void AddSound(string soundName, AudioClip soundData)
		{ if (Sounds.ContainsKey(soundName)) Sounds[soundName] = soundData; else Sounds.Add(soundName, soundData); }

		/// <summary> Plays a sound in a given position. </summary>
		/// <param name="soundName"> The sound name.</param>
		/// <param name="position"> The position of the audio source.</param>
		/// <param name="volume"> The sound volume. </param>
		public static void PlaySound(string soundName, Vector3 position = new Vector3(), float volume = 1.0f)
		{
			//Check the sound name string
			if(soundName == null || soundName.Length == 0) return;

			// If the Sound library does not exist yet, create it
			if (Sounds == null) Sounds = new Dictionary<string, AudioClip>();

			// Check if the sound is already in the library
			if (!Sounds.ContainsKey(soundName))
			{
				// Load the audio resource
				AudioClip resourceAudioClip = Resources.Load<AudioClip>(SoundsResourcePath + soundName);

				//Check the audio resource
				if (resourceAudioClip == null) { /*Debug.LogWarning("Sound Resource Null:" + soundName );*/ return;}

				// Add the sound resource to the library
				Sounds.Add(soundName, resourceAudioClip);
			}

			// If it does not exists yet, create the speaker object
			if (Speaker == null) {
				Speaker = new GameObject("Speaker");
				Speaker.transform.parent = Manager.Instance.transform;
				Speaker.AddComponent<AudioSource>();
			}

			// Position the speaker object
			Speaker.transform.position = position;

			// Check the sound library
			if (Sounds[soundName] == null) { /*Debug.LogWarning("Sound Null:" + soundName );*/ return; }

			// Play the audio
			Speaker.audio.PlayOneShot(Sounds[soundName], volume);
		}

		/// <summary> Check if a specifi sound is contained in the library. </summary>
		/// <param name="soundName"> The sound name. </param>
		public static bool SoundExists(string soundName) { return Sounds.ContainsKey(soundName); }

		public static byte[] GetSoundData(string soundName) 
		{
			// Make sure the audio clip actually exists
			if (!Sounds.ContainsKey(soundName)) return null;
			
			// Get the audio clip
			AudioClip clip = Sounds[soundName];	 

			// Get the sample data
			float[] samples = new float[clip.samples];
			clip.GetData(samples, 0);

			// Create buffers for the data conversion
			Int16[] intData = new Int16[samples.Length];
			Byte[] bytesData = new Byte[samples.Length * 2];
			
			// To convert float to Int16
			int rescaleFactor = 32767; 
			
			// Convert the sample data
			for (int i = 0; i<samples.Length; i++) {
				intData[i] = (short) (samples[i] * rescaleFactor);
				Byte[] byteArr = new Byte[2];
				byteArr = BitConverter.GetBytes(intData[i]);
				byteArr.CopyTo(bytesData, i * 2);
			}
			
			return bytesData;
		}

		public static void SaveSound(string soundName, string filePath)
		{
			// Make sure the audio clip actually exists
			if (!Sounds.ContainsKey(soundName)) return;

			// Get the audio clip
			AudioClip clip = Sounds[soundName];	 

			//Create a new file, ready to write on it
			FileStream fileStream = new FileStream(filePath, FileMode.Create);

			// Write the WAV heading (based on https://gist.github.com/darktable/2317063 )
			for(int i = 0; i < 44; i++) fileStream.WriteByte(new byte());
			fileStream.Seek(0, SeekOrigin.Begin);
			fileStream.Write(System.Text.Encoding.UTF8.GetBytes("RIFF"), 0, 4);						// Riff
			fileStream.Write(BitConverter.GetBytes(fileStream.Length - 8), 0, 4);					// Chunk size
			fileStream.Write(System.Text.Encoding.UTF8.GetBytes("WAVE"), 0, 4);						// Wave
			fileStream.Write(System.Text.Encoding.UTF8.GetBytes("fmt "), 0, 4);						// FMT
			fileStream.Write(BitConverter.GetBytes(16), 0, 4);										// Subchunk 1
			fileStream.Write(BitConverter.GetBytes(((UInt16)1)), 0, 2);								// Audio format
			fileStream.Write(BitConverter.GetBytes(clip.channels), 0, 2);							// Channels
			fileStream.Write(BitConverter.GetBytes(clip.frequency), 0, 4);							// Sample rate
			fileStream.Write(BitConverter.GetBytes(clip.frequency * clip.channels * 2), 0, 4);		// Byte rate
			fileStream.Write(BitConverter.GetBytes(((UInt16)((ushort) (clip.channels * 2)))), 0, 2);// Block align
			fileStream.Write(BitConverter.GetBytes((UInt16)(16)), 0, 2);							// Bits per sample
			fileStream.Write(System.Text.Encoding.UTF8.GetBytes("data"), 0, 4);						// Data String
			fileStream.Write(BitConverter.GetBytes(clip.samples * clip.channels * 2), 0, 4);		// Subchunk 2

			// Write the actual bytes
			Byte[] bytesData = GetSoundData(soundName);
			fileStream.Write(bytesData, 0, bytesData.Length);

			// Close the filestream
			fileStream.Close();
		}

		public static void Vibrate() { Handheld.Vibrate(); }
	}
}
