using UnityEngine;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Defines a selector object. </summary>
	public class Selector : MonoBehaviour
	{
		#region Private Fields
		
		/// <summary> The widget node. </summary>
		private Widget WidgetNode;
		
		/// <summary> The widget behaviour. </summary>
		private WidgetBehaviour Widget;
		
		/// <summary> The widget element. </summary>
		private GameObject WidgetObject;
		
		/// <summary> The parent widget node. </summary>
		private Widget ParentWidgetNode;
		
		/// <summary> The parent widget behaviour. </summary>
		private WidgetBehaviour ParentWidget;
		
		/// <summary> The parent widget element. </summary>
		private GameObject ParentWidgetObject;
		
		/// <summary> The current selection operation node. </summary>
		private Text CurrentSelectionOperation;
		
		private bool IsDefinitionPossible;
		
		private bool IsSelectionPossible;
		
		private float CurrentSelectorSize;
		
		private float InitialSelectorSize;
		
		private float MinimumSelectorSize;
		
		private float MaximumSelectorSize;
		
		private float SelectorSizeIncrement;
		
		private float SelectorSizeDecrement;
		
		/// <summary> The widget element. </summary>
		private Transform GeneratorObject;

		private Transform StructureObject;

		private Renderer[] Renderers;
		
		#endregion
		
		#region Private Methods
		
		/// <summary> Initialize the basic manipulator properties. </summary>
		private void Start()
		{
			// Get the parent widget
			ParentWidget = transform.parent.GetComponent<WidgetBehaviour>();
			ParentWidgetObject = ParentWidget.gameObject;
			
			CurrentSelectionOperation = Manager.GetResource(null, "CurrentOperation");
			
			CurrentSelectorSize = 0f;
			SelectorSizeIncrement = 10f;
			SelectorSizeDecrement = -2f;
			MinimumSelectorSize = 0f;
			MaximumSelectorSize = 1.5f;
			IsDefinitionPossible = true;
			
			foreach (Transform child in transform.parent.parent) 
			{
				if (child.name.Contains("Generator")) GeneratorObject = child;
				if (child.name.Contains("Structure")) StructureObject = child;
			}

			Renderers = GetComponentsInChildren<Renderer>();
		}
		
		/// <summary>  </summary>
		private void Update()
		{
			if (CurrentSelectionOperation.Value != "SelectionArea") { transform.localScale = Vector3.zero; return; }
			
			//if (ParentWidget.Enabled == false) { CurrentSelectorSize = 0f; return; }



			if (InputManager.CurrentAction != InputManager.Action.None && InputManager.Hit.collider != null
			    && InputManager.Hit.collider.gameObject != null && IsDefinitionPossible)
			{
				this.transform.position = InputManager.HitPosition;
				
				if(CurrentSelectorSize < MinimumSelectorSize) CurrentSelectorSize = MinimumSelectorSize;
				if(CurrentSelectorSize < MaximumSelectorSize) CurrentSelectorSize += SelectorSizeIncrement * Time.deltaTime;
				if(CurrentSelectorSize >= MaximumSelectorSize) CurrentSelectorSize = MaximumSelectorSize;
				
				IsSelectionPossible = true;
			}
			else
			{
				if (IsSelectionPossible && CurrentSelectorSize == MaximumSelectorSize) 
				{
					float explosionSize = CurrentSelectorSize / 5.0f;
					
					Explosion.Create(ParentWidgetObject.transform, transform.position, explosionSize, 1, UnityEngine.Color.white);
					
					// Play the appropiate sound
					OutputManager.PlaySound("SelectionArea", transform.position);
					
					// Check all the targets
					foreach (Transform target in GeneratorObject)
					{
						// If the targets are too close to the explosion, mark them as selected
						if (Vector3.Distance(this.transform.position, target.position) < (explosionSize /2.0f))
						{ 
							Target selectedTarget = target.GetComponent<Target>();
							if (selectedTarget != null) selectedTarget.React("selected");
						}
					}
					
					// Check the distance to the structure
					if (Vector3.Distance(this.transform.position, StructureObject.position) < (explosionSize))
						MetricsManager.AddError(MetricsManager.TaskId.UC3C);
					
					IsSelectionPossible = false;
				}
				
				if(CurrentSelectorSize > MinimumSelectorSize) CurrentSelectorSize += SelectorSizeDecrement * Time.deltaTime;
				if(CurrentSelectorSize <= MinimumSelectorSize) CurrentSelectorSize = MinimumSelectorSize;
				
				IsDefinitionPossible = (CurrentSelectorSize == MinimumSelectorSize);
			}
			
			this.transform.localScale = new Vector3(CurrentSelectorSize, CurrentSelectorSize, CurrentSelectorSize);

			foreach (Renderer r in Renderers)
			{
				r.enabled = (CurrentSelectorSize > 0);
				r.material.color = (CurrentSelectorSize == MaximumSelectorSize)? 
					new UnityEngine.Color(1,1,1) : new UnityEngine.Color(0.5f,0.5f,0.5f);
			}
		}
		
		#endregion
	}
}
