﻿using UnityEngine;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Defines the behaviour of a slelection target. </summary>
	public class Structure : MonoBehaviour
	{
		#region Private Fields
		
		/// <summary> The widget behavior. </summary>
		private WidgetBehaviour Widget;
		
		/// <summary> The current selection operation node. </summary>
		private Text CurrentSelectionOperation;

		/// <summary> The structural integrity value. </summary>
		private Text StructuralIntegrity;

		/// <summary> The color used to represent the damage. </summary>
		private UnityEngine.Color DamageColor;

		/// <summary> The time the damage is represented. </summary>
		private float DamageTime;

		/// <summary> The current time the damage is represented. </summary>
		private float DamageCurrentTime;

		/// <summary> The original color of the object. </summary>
		private UnityEngine.Color DamageOriginalColor;

		private Text UseCase3Errors;

		private Text UseCase3State;

		#endregion
		
		#region Public Fields

		/// <summary> The damage inflected by simple selection operations. </summary>
		public int SelectionSimpleDamage = 0;

		/// <summary> The damage inflected by multiple selection operations. </summary>
		public int SelectionMultipleDamage = 0;

		/// <summary> The damage inflected by area selection operations. </summary>
		public int SelectionAreaDamage = 0;

		#endregion
		
		#region Private Methods
		
		/// <summary> Initialize the instance. </summary>
		private void Start ()
		{
			// Get the widget
			Widget = GetComponentInParent<WidgetBehaviour>();

			CurrentSelectionOperation = Manager.GetResource(null, "CurrentOperation");
			StructuralIntegrity = Manager.GetResource(Widget.WidgetNode, "Integrity");
			UseCase3State = Manager.GetResource(null, "UseCase3State");
			DamageTime = 0.0f;
		}
		
		
		/// <summary> Called once per frame. </summary>
		private void LateUpdate ()
		{
			if (StructuralIntegrity != null && int.Parse(StructuralIntegrity.Value) <= 0 && UseCase3State == "0")
			{
				// Create a Big explosion
				Explosion.Create(this.transform.parent.parent, Vector3.zero, 0.6f, 3, UnityEngine.Color.white);

				// Play the appropiate sound
				OutputManager.PlaySound("SelectionLost", transform.position);

				// Change the game state
				UseCase3State.Update("1");
			}

			// Apply the damage effect
			if (DamageTime != 0.0f)
			{
				// Add the current delta time
				DamageCurrentTime += Time.deltaTime;

				// Clamp the effct
				if (DamageCurrentTime > DamageTime) DamageCurrentTime = DamageTime;

				// Interpolate the colors
				this.renderer.material.color = UnityEngine.Color.Lerp(DamageColor, DamageOriginalColor, 
				                                                DamageCurrentTime / DamageTime);

				// Check the end
				if (DamageCurrentTime == DamageTime) DamageCurrentTime = DamageTime = 0.0f;
			}
		}
		
		
		/// <summary> Called when the object is clicked upon. </summary>
		/// <param name="hit"> The hit data.</param>
		private void OnClicked(RaycastHit hit)
		{
			// FIX Make absolutely sure that we are in the right section
			if (Root.Instance.Platforms[0].Resources[0].Texts["Section"] != "UseCase3Execution") return;

			// Make sure the current operation is simple selection
			if (CurrentSelectionOperation != "SelectionSimple") return;
			
			// Create a small explosion to show where the object is hit
			Explosion.Create(Widget.Parent.transform, hit.point, 0.01f, 1, UnityEngine.Color.white);
			
			// Play the appropiate sound
			OutputManager.PlaySound("SelectionSimple", hit.point);
			
			// Perform the selected action
			React("selected");
		}
		
		/// <summary> Called when the object is being pressed on. </summary>
		/// <param name="hit"> The hit data. </param>
		private void OnCrossed(RaycastHit hit)
		{
			// FIX Make absolutely sure that we are in the right section
			if (Root.Instance.Platforms[0].Resources[0].Texts["Section"] != "UseCase3Execution") return;

			// Make sure the current operation is multiple selection
			if (CurrentSelectionOperation != "SelectionMultiple") return;
			
			// Create a small explosion to show where the object is hit
			Explosion.Create(Widget.Parent.transform, hit.point, 0.01f, 1, UnityEngine.Color.white);
			
			// Play the appropiate sound
			OutputManager.PlaySound("SelectionMultiple", hit.point);
			
			// Perform the selected action
			React("selected");
		}

		#endregion
		
		#region Public Methods
		
		/// <summary> Reacts to an specified action. </summary>
		/// <param name="actionName"> The name of the action. </param>
		public void React(string actionName)
		{
			// Make sure that the widget is enabled
			if (!Widget.Enabled) return;

			// Check for selection errors
			if (actionName != "selected" || StructuralIntegrity == null) return;
			int damage = 0;
			switch (CurrentSelectionOperation.Value)
			{
			case "SelectionSimple":
				damage = SelectionSimpleDamage; MetricsManager.AddError(MetricsManager.TaskId.UC3A); break;
			case "SelectionMultiple":
				damage = SelectionMultipleDamage; MetricsManager.AddError(MetricsManager.TaskId.UC3B); break;
			case "SelectionArea":
				damage = SelectionAreaDamage; MetricsManager.AddError(MetricsManager.TaskId.UC3C); break;
			}

			if (damage != 0)
			{
				// Play a sound
				OutputManager.PlaySound("SelectionError", transform.position);

				// Update the structural integrity and user errors value
				StructuralIntegrity.Update(((int.Parse(StructuralIntegrity.Value)) - damage).ToString());

				// Set the damage effect fields
				DamageCurrentTime = 0.0f; DamageTime = damage / 10.0f;
				DamageOriginalColor = renderer.material.color;
				DamageColor = new UnityEngine.Color(1.0f - DamageOriginalColor.r, 1.0f - DamageOriginalColor.g,
				                                    1.0f - DamageOriginalColor.b, 1.0f - DamageOriginalColor.a);
			}
		}

		#endregion
	}
}