﻿using UnityEngine;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Defines the behaviour of a slelection target. </summary>
	public class Target : MonoBehaviour
	{
		#region Private Fields
		
		/// <summary> The widget behavior. </summary>
		private WidgetBehaviour Widget;

		/// <summary> The current selection operation node. </summary>
		private Text CurrentSelectionOperation;

		/// <summary> The current life time. </summary>
		private float CurrentLifeTime = 0.0f;

		private Vector3 StartPosition;

		private bool DestroyRequested;
		#endregion

		#region Public Fields

		/// <summary> Total life time. </summary>
		public float LifeTime = 0.0f;

		/// <summary> The translation speed. </summary>
		public float Speed = 0.0f;

		#endregion

		#region Private Methods
		
		/// <summary> Initialize the instance. </summary>
		private void Start ()
		{
			// Get the widget
			Widget = GetComponent<WidgetBehaviour>();
			if (Widget ==null) Widget = this.GetComponentInParent<WidgetBehaviour>();

			if (LifeTime > 0)
			{ 
//				Widget.transform.localScale = new Vector3(0.001f, 0.001f, 0.001f);
				Widget.Enabled = true;
				StartPosition = this.transform.localPosition;
			}

			CurrentSelectionOperation = Manager.GetResource(null, "CurrentOperation");
			DestroyRequested = false;
		}
		
		
		/// <summary> Called once per frame. </summary>
		private void LateUpdate ()
		{
			// If the target is no longer visible, destroy it to avoid further problems
			if(DestroyRequested && !animation.isPlaying) { Widget.WidgetNode.Destroy(); return; }

			if(this.transform.localPosition.magnitude < 0.11f) React("structureReached");


			// Make sure that the widget is enabled
			if(!Widget.Enabled) return;

			// If the current selection operation is area selection, disable the colliders
			bool areaSelectionEnabled = (CurrentSelectionOperation != "SelectionArea");
			if(this.collider.enabled != areaSelectionEnabled) this.collider.enabled = areaSelectionEnabled; 

			if (LifeTime > 0)
			{
				CurrentLifeTime += Time.deltaTime;
				if (CurrentLifeTime >= LifeTime) Widget.Enabled = false;
				this.transform.localPosition = StartPosition * (((LifeTime - CurrentLifeTime) / LifeTime) + Speed);
				this.transform.LookAt(Vector3.zero);
			}
		}

		/// <summary> Called when the object is clicked upon. </summary>
		/// <param name="hit"> The hit data.</param>
		private void OnClicked(RaycastHit hit)
		{
			// Make sure the current operation is simple selection
			if (CurrentSelectionOperation != "SelectionSimple") return;

			// Create a small explosion to show where the object is hit
			Explosion.Create(Widget.Parent.transform, hit.point, 0.01f, 1, UnityEngine.Color.white);

			// Play the appropiate sound
			OutputManager.PlaySound("SelectionSimple", hit.point);

			// Perform the selected action
			React("selected");
		}

		/// <summary> Called when the object is being pressed on. </summary>
		/// <param name="hit"> The hit data. </param>
		private void OnCrossed(RaycastHit hit)
		{
			// Make sure the current operation is multiple selection
			if (CurrentSelectionOperation != "SelectionMultiple") return;
			
			// Create a small explosion to show where the object is hit
			Explosion.Create(Widget.Parent.transform, hit.point, 0.01f, 1, UnityEngine.Color.white);
			
			// Play the appropiate sound
			OutputManager.PlaySound("SelectionMultiple", hit.point);
			
			// Perform the selected action
			React("selected");
		}

		#endregion
		
		#region Public Methods

		/// <summary> Reacts to an specified action. </summary>
		/// <param name="actionName"> The name of the action. </param>
		public void React(string actionName)
		{
			// Make sure that the widget is enabled
			if (!Widget.Enabled) return;

			// Disable the widget
			if (LifeTime > 0) { Widget.Enabled = false; DestroyRequested = true; }

			// Create an explosion
			Explosion.Create(transform.parent, transform.localPosition, 0.1f, 1.0f, UnityEngine.Color.red);

			// Play the appropiate sound
			if (audio != null) audio.Play ();

			// Call the parent node
			Widget.Parent.React(actionName);
		}

		#endregion
	}
}