using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Avatar personalization keyboard. </summary>
	public class KeyboardAvatar : Keyboard
	{
		#region Private Fields

		private const float InitialHeight = 1.7f;

		#endregion

		#region Private Fields

		/// <summary> The model values. </summary>
		private string[] ModelValues;

		/// <summary> The model index. </summary>
		private int ModelIndex;

		/// <summary> The model Text Node. </summary>
		private Text ModelValue;

		/// <summary> The height values. </summary>
		private string[] HeightValues;
		
		/// <summary> The height index. </summary>
		private int HeightIndex;

		/// <summary> The height Text Node. </summary>
		private Text HeightValue;

		/// <summary> The figure values. </summary>
		private string[] FigureValues;
		
		/// <summary> The figure index. </summary>
		private int FigureIndex;

		/// <summary> The figure Text Node. </summary>
		private Text FigureValue;

		/// <summary> The avatar GameObject. </summary>
		private GameObject Avatar;

		/// <summary> The different avatar components. </summary>
		private Dictionary<string, GameObject> Components;

		/// <summary> The initial position of the head. </summary>
		private Vector3 HeadInitialPosition;

		/// <summary> The initial position of the head. </summary>
		private Quaternion HeadInitialRotation;

		/// <summary> The last Gender index. </summary>
		private bool ModelAlreadySelected;

//		/// <summary> The last time that the avatar saluted. </summary>
//		private float SaluteLastTime;
//
//		/// <summary> The frenquency of avatar salutes. </summary>
//		private float SaluteFrequency;

		#endregion
		
		#region Private Methods
		
		/// <summary> Initializes the keyboard. </summary>
		private void Start()
		{
			// Initialize the basic elements
			Initialize();

			// Define the different options and their indexes
			System.StringSplitOptions noNulls = System.StringSplitOptions.RemoveEmptyEntries;
			string[] separators = new string[]{",", " "};
			ModelValues = Manager.GetResource(ParentWidgetNode, "AvatarModelValues").Value.Split(separators, noNulls); 
			HeightValues = Manager.GetResource(ParentWidgetNode, "AvatarHeightValues").Value.Split(separators, noNulls); 
			FigureValues = Manager.GetResource(ParentWidgetNode, "AvatarFigureValues").Value.Split(separators, noNulls); 
			ModelValue = Manager.GetResource(ParentWidgetNode, "AvatarModelValue"); 
			HeightValue = Manager.GetResource(ParentWidgetNode, "AvatarHeightValue"); 
			FigureValue = Manager.GetResource(ParentWidgetNode, "AvatarFigureValue"); 
			for (int i = 0; i < ModelValues.Length; i++) if (ModelValue.Value == ModelValues[i]) ModelIndex = i;
			for (int i = 0; i < HeightValues.Length; i++) if (HeightValue.Value == HeightValues[i]) HeightIndex = i;
			for (int i = 0; i < FigureValues.Length; i++) if (FigureValue.Value == FigureValues[i]) FigureIndex = i;


			// Update the text meshes
			Texts["ModelLabel"].text = Manager.GetResource(ParentWidgetNode, "AvatarModelLabel");
			Texts["HeightLabel"].text = Manager.GetResource(ParentWidgetNode, "AvatarHeightLabel");
			Texts["FigureLabel"].text = Manager.GetResource(ParentWidgetNode, "AvatarFigureLabel");
			Texts["Instructions"].text = Manager.GetResource(ParentWidgetNode, "Instructions").Value.Replace("//", "\n");
			Texts["ModelValue"].text = ModelValues[ModelIndex];
			Texts["HeightValue"].text = HeightValues[HeightIndex];
			Texts["FigureValue"].text = FigureValues[FigureIndex];

			// Get the avatar GameObject and the internal models
			Avatar = this.GetComponentInChildren<Animator>().gameObject;
			Components = new Dictionary<string, GameObject>();
			foreach(Transform child in Avatar.transform) Components.Add(child.name, child.gameObject);

			// Store the initial position and rotation of the head
			HeadInitialPosition = Components["Head"].transform.localPosition;
			HeadInitialRotation = Components["Head"].transform.localRotation;

			// Make sure that the avatar is properly updated
			ModelAlreadySelected = false;
			Root.Instance.Users[0].Personas[0].Gender.UpdatePerformed += UpdateAvatarValues;
			Root.Instance.Users[0].Personas[0].Laterality.UpdatePerformed += UpdateAvatarValues;
			ParentWidget.SelectedStateChanged += UpdateAvatar;

			// Start with the salute animation and wait 30 seconds until the next animation
			//SaluteLastTime = 28.0f; SaluteFrequency = 30.0f;

			
			// Detect when the parent widget is selected
			ParentWidget.SelectedStateChanged += this.OnWidgetSelected;
		}
		
		
		/// <summary> Detects when the keyboard is loaded. </summary>
		private void OnWidgetSelected (WidgetBehaviour widget)
		{ if (widget.Selected && !TextNode.IsEmpty) MetricsManager.AddError(MetricsManager.TaskId.UC1C); }


		/// <summary> Called after the frame drawing. </summary>
		private void LateUpdate()
		{
			if (!ParentWidget.Selected || !AugmentedReality.Initialized) return;

			Transform head = Components["Head"].transform;
			head.LookAt(AugmentedReality.CameraObject.transform);

			while(Quaternion.Angle(head.localRotation, HeadInitialRotation) > 60.0f)
				head.localRotation = Quaternion.Lerp(HeadInitialRotation, head.localRotation, 0.95f);

			MetricsManager.AddTime(MetricsManager.TaskId.UC1C, Time.deltaTime);
		}

		/// <summary> Updates the avatar values. </summary>
		private void UpdateAvatarValues(Node node)
		{
			// Check if the model has already been selected
			if (!ModelAlreadySelected)
			{
				// Get the model index
				ModelIndex = (Root.Instance.Users[0].Personas[0].Gender.Value == 1)? 1 : 0; 

				// Update the text
				Texts["ModelValue"].text = ModelValues[ModelIndex];
			}

			// Update the avatar
			UpdateAvatar(null);
		}

		/// <summary> Updates the avatar model. </summary>
		private void UpdateAvatar(WidgetBehaviour widget)
		{
			// Select the right model
			string currentModel = ((ModelIndex == 0)? "Male" : "Female") + "00" + (FigureIndex + 1).ToString();
			foreach(GameObject model in Components.Values)
				if (model.name.Contains("ale"))	model.renderer.enabled = (model.name == currentModel);
			
			// Update the height value
			float laterality =  (Root.Instance.Users[0].Personas[0].Laterality != 1)? 1.0f : -1.0f;
			float height = 1.5f + (0.05f * HeightIndex);
			Components["Skeleton"].transform.localScale = new Vector3(laterality, 1.0f, 1.0f) * height / InitialHeight;
			Components["Head"].transform.localPosition = new Vector3(
				HeadInitialPosition.x,  HeadInitialPosition.y * height / InitialHeight, HeadInitialPosition.z);

			//
			if(CameraImage.FinalImage != null)
			{
				GameObject face = GameObject.Find("Face");
				face.renderer.material.SetTexture("_Texture2", CameraImage.FinalImage);
				face.renderer.material.SetTextureScale("_Texture2", new Vector2(0.9f,1.1f));
				face.renderer.material.SetTextureOffset("_Texture2",  new Vector2(0f,0f));
			}
		}

		#endregion
		
		#region Public Methods
		
		/// <summary> Called upon key pressing. </summary>
		/// <param name="key"> The pressed Key. </param>
		public override void KeyPressed(Key key)
		{
			// Analyze the given key name
			switch (key.name)
			{
			case "Key_11":
				if (ModelIndex > 0) ModelIndex--; else ModelIndex = ModelValues.Length - 1;
				Texts["ModelValue"].text = ModelValues[ModelIndex];
				Manager.GetResource(ParentWidgetNode, "AvatarModelValue").Update(ModelValues[ModelIndex]);
				ModelAlreadySelected = true;
				break;
			case "Key_12":
				if (ModelIndex < ModelValues.Length - 1) ModelIndex++; else ModelIndex = 0;
				Texts["ModelValue"].text = ModelValues[ModelIndex];
				Manager.GetResource(ParentWidgetNode, "AvatarModelValue").Update(ModelValues[ModelIndex]);
				ModelAlreadySelected = true;
				break;
			case "Key_21":
				if (HeightIndex > 0) HeightIndex--; else HeightIndex = HeightValues.Length - 1;
				Texts["HeightValue"].text = HeightValues[HeightIndex];
				Manager.GetResource(ParentWidgetNode, "AvatarHeightValue").Update(HeightValues[ModelIndex]);
				break;
			case "Key_22":
				if (HeightIndex < HeightValues.Length - 1) HeightIndex++; else HeightIndex = 0;
				Texts["HeightValue"].text = HeightValues[HeightIndex];
				Manager.GetResource(ParentWidgetNode, "AvatarHeightValue").Update(HeightValues[ModelIndex]);
				break;
			case "Key_31":
				if (FigureIndex > 0) FigureIndex--; else FigureIndex = FigureValues.Length - 1;
				Texts["FigureValue"].text = FigureValues[FigureIndex];
				Manager.GetResource(ParentWidgetNode, "AvatarFigureValue").Update(FigureValues[ModelIndex]);
				break;
			case "Key_32":
				if (FigureIndex < FigureValues.Length - 1) FigureIndex++; else FigureIndex = 0;
				Texts["FigureValue"].text = FigureValues[FigureIndex];
				Manager.GetResource(ParentWidgetNode, "AvatarFigureValue").Update(FigureValues[ModelIndex]);
				break;
			case "Key_41": Avatar.transform.parent.Rotate(0.0f,+30.0f,0.0f, Space.Self); return;
			case "Key_42": Avatar.transform.parent.Rotate(0.0f,-30.0f,0.0f, Space.Self); return;
			case "AvatarBase":
				Texts["Instructions"].text = "";
				Avatar.animation.Play("Salute"); 
				if (OutputManager.SoundExists("Salute")) OutputManager.PlaySound("Salute", Avatar.transform.position); 
				// SaluteLastTime = 0;
				return;
			case "Key_OK": 
				ParentWidget.Selected = false;
				TextNode.Update(Manager.GetResource(ParentWidgetNode, "Obtained").Value);
				return;
			}

			// Update the avatar model
			UpdateAvatar(null);
		}
		
		#endregion
	}
}

