﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Multiple option selection keyboard. </summary>
	public class KeyboardMultipleOptions : Keyboard
	{
		#region Private Class

		/// <summary> The options data structure. </summary>
		private class Option { public string Label; public string[] Values; public int ValueIndex; public Text Value; }
		
		#endregion
		
		#region Private Fields

		/// <summary> The different options.</summary>
		private Option[] Options;

		#endregion
		
		#region Private Methods
		
		/// <summary> Initializes the keyboard. </summary>
		private void Start()
		{
			// Initialize the basic elements
			Initialize();

			// Store the option data
			Options = new Option[3];
			for (int optionIndex = 0; optionIndex < 3; optionIndex++)
			{
				Option option = Options[optionIndex] = new Option();
				option.Label = Manager.GetResource(ParentWidgetNode, "Option" + optionIndex + "Label").Value;
				option.Values = Manager.GetResource(ParentWidgetNode, "Option" + optionIndex + "Options")
					.Value.Split(new string[]{","}, System.StringSplitOptions.RemoveEmptyEntries);
				option.Value = Manager.GetResource(ParentWidgetNode, "Option" + optionIndex + "Value");
				option.ValueIndex = 0; 
				
				// Initialize the text meshes
				Texts["Option" + optionIndex + "Label"].text = option.Label;
				Texts["Option" + optionIndex + "Value"].text = option.Values[0];
			}

			// Detect when the parent widget is selected
			ParentWidget.SelectedStateChanged += this.OnWidgetSelected;
		}


		/// <summary> Detects when the keyboard is loaded. </summary>
		private void OnWidgetSelected (WidgetBehaviour widget)
		{ if (widget.Selected && !TextNode.IsEmpty) MetricsManager.AddError(MetricsManager.TaskId.UC1B); }


		/// <summary> Called after the frame drawing. </summary>
		private void LateUpdate()
		{	if (ParentWidget.Selected) MetricsManager.AddTime(MetricsManager.TaskId.UC1B, Time.deltaTime); }

		#endregion
		
		#region Public Methods
		
		/// <summary> Called upon key pressing. </summary>
		/// <param name="key"> The pressed Key. </param>
		public override void KeyPressed(Key key)
		{
			// Check the given value
			if (key.Values == null || key.Values[key.CurrentIndex] == null) return;
			
			// Get the key properties
			string keyName = key.name;

			// Analyze the given key name
			int optionIndex = 0, valueIndexIncrement = 0;
			switch (keyName)
			{
			case "Key_01": ParentWidget.Selected = false; break;
			case "Key_11": optionIndex = 0; valueIndexIncrement = -1; break;
			case "Key_12": optionIndex = 0; valueIndexIncrement = +1; break;
			case "Key_21": optionIndex = 1; valueIndexIncrement = -1; break;
			case "Key_22": optionIndex = 1; valueIndexIncrement = +1; break;
			case "Key_31": optionIndex = 2; valueIndexIncrement = -1; break;
			case "Key_32": optionIndex = 2; valueIndexIncrement = +1; break;
			}


			// Get the new index
			Option option = Options[optionIndex];
			if (valueIndexIncrement != 0)
			{
				if (valueIndexIncrement > 0)
				{ if (option.ValueIndex < option.Values.Length -1) option.ValueIndex++; else option.ValueIndex = 0; }
				else { if (option.ValueIndex > 0) option.ValueIndex--; else option.ValueIndex = option.Values.Length -1; }
				
				// Update the value node
				option.Value.Update(option.ValueIndex.ToString());
				
				// Update the textmesh
				Texts["Option" + optionIndex + "Value"].text = option.Values[option.ValueIndex];
			}

			// FIX: Store the values in the persona node
			switch (optionIndex)
			{
			case 0: Root.Instance.Users[0].Personas[0].Laterality.Update(Options[0].ValueIndex); break;
			case 1: Root.Instance.Users[0].Personas[0].Vision.Update(Options[1].ValueIndex); break;
			case 2: Root.Instance.Users[0].Personas[0].ColorBlindness.Update(Options[2].ValueIndex); break;
			}

			// Update the text node
			string textValue = "";
			for (optionIndex = 0; optionIndex < 3; optionIndex++)
			{
				// Get the current option
				option = Options[optionIndex];
				
				// If the option is not the dafault one, add it to the string
				if (option.ValueIndex != 0)
				{
					// If it is not the first element, and a conjuction
					if (textValue.Length > 0) textValue += " y ";
					
					// Add the value to the message
					textValue += option.Values[option.ValueIndex];
					
					// If it is too complex, simplify the shown message
					if (textValue.Length > 20)
					{ textValue = Manager.GetResource(ParentWidgetNode, "ComplexValue").Value; break; }
				}
			}
			if (textValue == "") textValue =  Manager.GetResource(ParentWidgetNode, "DefaultValue").Value;

			TextNode.Update(textValue);
		}
		
		#endregion
	}
}

