using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Multiple selection keyboard. </summary>
	public class KeyboardMultipleSelection : Keyboard
	{
		#region Private Fields

		/// <summary> The list of strings with the different options that the user can select. </summary>
		private List<string> Options;

		/// <summary> Inidcates if a given option is enabled or not. </summary>
		private bool[] Values;

		#endregion
		
		#region Private Methods
		
		/// <summary> Initializes the keyboard. </summary>
		private void Start()
		{
			// Initialize the basic elements
			Initialize();

			Options = new List<string>(Manager.GetResource(ParentWidgetNode, "Options")
				.Value.Split(new string[]{","}, System.StringSplitOptions.RemoveEmptyEntries));

			for (int i = 0; i < 9; i++) { Keys[i].Values[0] = Options[i]; Keys[i].ChangeIndex(0, true); }

			Values = new bool[9];
			for (int i = 0; i < 9; i++) { Values[i] = false;}

			// Detect when the parent widget is selected
			ParentWidget.SelectedStateChanged += this.OnWidgetSelected;
		}
		
		
		/// <summary> Detects when the keyboard is loaded. </summary>
		private void OnWidgetSelected (WidgetBehaviour widget)
		{ if (widget.Selected && !TextNode.IsEmpty) MetricsManager.AddError(MetricsManager.TaskId.UC1B); }


		/// <summary> Updates this instance once per frame. </summary>
		protected override void Update()
		{
			base.Update();

			if (!ParentWidget.Selected) return;

			UnityEngine.Color keyColor = Keys[0].transform.parent.renderer.material.color;
			for (int i = 0; i < 9; i++) Keys[i].renderer.material.color = (Values[i])? keyColor * 2.0f : keyColor;
		}

		/// <summary> Called after the frame drawing. </summary>
		private void LateUpdate()
		{	if (ParentWidget.Selected) MetricsManager.AddTime(MetricsManager.TaskId.UC1B, Time.deltaTime); }

		#endregion
		
		#region Public Methods
		
		/// <summary> Called upon key pressing. </summary>
		/// <param name="key"> The pressed Key. </param>
		public override void KeyPressed(Key key)
		{
			// Check the given value
			if (key.Values == null || key.Values[key.CurrentIndex] == null) return;

			// Get the text value
			string keyValue = key.Values[key.CurrentIndex];

			// Analyze the given value
			switch (keyValue)
			{
			case "OK": ParentWidget.Selected = false; break;
			default: int index = Options.IndexOf(keyValue);	Values[index] = !Values[index];	break;
			}

			// Update the text node
			string textValue = "";
			for (int optionIndex = 0; optionIndex < 9; optionIndex++)
			{
				// If the option is not the dafault one, add it to the string
				if (Values[optionIndex])
				{
					// If it is not the first element, and a conjuction
					if (textValue.Length > 0) textValue += " y ";
					
					// Add the option value to the message
					textValue += Options[optionIndex];
					
					// If it is too complex, simplify the shown message
					if (textValue.Length > 20)
					{ textValue = Manager.GetResource(ParentWidgetNode, "ComplexValue").Value; break; }
				}
			}
			if (textValue == "") textValue =  Manager.GetResource(ParentWidgetNode, "DefaultValue").Value;
			TextNode.Update(textValue);

			// FIX: Store the values in the persona node
			Root.Instance.Users[0].Personas[0].Interests.Update(textValue);
		}

		
		#endregion
	}
}

