﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Single option selection keyboard. </summary>
	public class KeyboardSingleOption : Keyboard
	{
		#region Private Fields

		/// <summary> The list of strings with the different options that the user can select. </summary>
		private List<string> Options;

		/// <summary> The total number of pages. </summary>
		private int TotalPages;

		/// <summary> The current page. </summary>
		private int CurrentPage;

		/// <summary> The current index. </summary>
		private int CurrentIndex;

		#endregion
		
		#region Private Methods
		
		/// <summary> Initializes the keyboard. </summary>
		private void Start()
		{
			// Initialize the basic elements
			Initialize();

			// Store the option data
			Options = new List<string>(Manager.GetResource(ParentWidgetNode, "Options")
				.Value.Split(new string[]{","}, System.StringSplitOptions.RemoveEmptyEntries));

			// Update the key texts
			TotalPages = Options.Count / 3; CurrentPage = 0;
			string [][] values = new string[3][];
			for (int valueIndex = 0; valueIndex < 3; valueIndex++)
			{
				values[valueIndex] = new string[TotalPages];
				for (int pageIndex = 0; pageIndex < TotalPages; pageIndex++)
				{
					int currentIndex = (pageIndex * 3) + valueIndex;
					if (currentIndex < Options.Count) values[valueIndex][pageIndex] = Options[currentIndex];
				}
			}
			// Update the page numbers
			string [] pageNumbers = new string[TotalPages];
			for (int pageIndex = 0; pageIndex < TotalPages; pageIndex++)
				pageNumbers[pageIndex] = string.Format("{0}/{1}", pageIndex+1, TotalPages);

			foreach(Key key in Keys)
			{
				switch (key.name)
				{
				case "Key_11": key.Values = values[0]; key.ChangeIndex(0, true); break;
				case "Key_21": key.Values = values[1]; key.ChangeIndex(0, true); break;
				case "Key_31": key.Values = values[2]; key.ChangeIndex(0, true); break;
				case "Key_12": key.ChangeIndex(0, true); break;
				case "Key_22": key.ChangeIndex(0, true); break;
				case "PageNumber": key.Values = pageNumbers; key.ChangeIndex(0, true); break;
				}
			}

			// Detect when the parent widget is selected
			ParentWidget.SelectedStateChanged += this.OnWidgetSelected;
		}
		
		
		/// <summary> Detects when the keyboard is loaded. </summary>
		private void OnWidgetSelected (WidgetBehaviour widget)
		{ if (widget.Selected && !TextNode.IsEmpty) MetricsManager.AddError(MetricsManager.TaskId.UC1B); }


		/// <summary> Called after the frame drawing. </summary>
		private void LateUpdate()
		{	if (ParentWidget.Selected) MetricsManager.AddTime(MetricsManager.TaskId.UC1B, Time.deltaTime); }

		#endregion
		
		#region Public Methods
		
		/// <summary> Called upon key pressing. </summary>
		/// <param name="key"> The pressed Key. </param>
		public override void KeyPressed(Key key)
		{
			// Check the given value
			if (key.Values == null || key.Values[key.CurrentIndex] == null) return;
			
			// Get the key properties
			string keyName = key.name;

			// Analyze the given key name
			switch (keyName)
			{
			case "Key_01": ParentWidget.Selected = false; return;
			case "Key_11": CurrentIndex = 0; TextNode.Update(key.Values[CurrentPage]); ParentWidget.Selected = false; break;
			case "Key_21": CurrentIndex = 1; TextNode.Update(key.Values[CurrentPage]); ParentWidget.Selected = false; break;
			case "Key_31": CurrentIndex = 2; TextNode.Update(key.Values[CurrentPage]); ParentWidget.Selected = false; break;
			case "Key_12": CurrentPage--; break;
			case "Key_32": CurrentPage++; break;
			}

			// Update the texts
			if (CurrentPage < 0) CurrentPage = TotalPages - 1;
			else if (CurrentPage >= TotalPages) CurrentPage = 0;

			foreach(Key k in Keys)
				if (k.name == "Key_11" || k.name == "Key_21" || k.name == "Key_31" || k.name == "PageNumber") 
					k.ChangeIndex(CurrentPage, true);


			// FIX: Store the values in the persona node
			Root.Instance.Users[0].Personas[0].Gender.Update((CurrentPage * 3) + CurrentIndex);
		}
		
		#endregion
	}
}

