﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> Sound recording keyboard. </summary>
	public class KeyboardSound : Keyboard
	{
		#region Private Fields

		/// <summary> The maximum length of the sound (in seconds). </summary>
		private int SoundLength = 3;

		/// <summary> The key that activates the sound recording. </summary>
		private Key KeyRecord;

		/// <summary> The key that activates the sound replay. </summary>
		private Key KeyPlay;

		#endregion
		
		#region Private Methods
		
		/// <summary> Initializes the keyboard. </summary>
		private void Start()
		{
			// Initialize the basic elements
			Initialize();

			// Set the text values
			Texts["Instructions"].text = Manager.GetResource(ParentWidgetNode, "Instructions");
			string[] KeyTexts = Manager.GetResource(ParentWidgetNode, "Texts").Value.Split(new string[]{","}, 
				System.StringSplitOptions.RemoveEmptyEntries);

			// Set the key values
			foreach(Key key in Keys)
			{
				switch (key.name)
				{
				case "Key_Record":
					KeyRecord = key;
					key.Values = new string[]{KeyTexts[0], KeyTexts[1], null};
					key.ChangeIndex(0, true);
					break;
				case "Key_Play":
					KeyPlay = key;
					key.Values = new string[]{null, KeyTexts[2], KeyTexts[3]};
					key.transform.localScale = Vector3.one * 0.001f;
					break;
				}
			}
			
			// Detect when the parent widget is selected
			ParentWidget.SelectedStateChanged += this.OnWidgetSelected;
		}
		
		
		/// <summary> Detects when the keyboard is loaded. </summary>
		private void OnWidgetSelected (WidgetBehaviour widget)
		{ if (widget.Selected && !TextNode.IsEmpty) MetricsManager.AddError(MetricsManager.TaskId.UC1C); }


		/// <summary> The sound has been recorded and is ready to play. </summary>
		private void SoundReady()
		{
			// Update the keys
			KeyRecord.ChangeIndex(0); KeyPlay.ChangeIndex(1);

			// Update the node text
			TextNode.Update(Manager.GetResource(ParentWidgetNode, "Obtained"));
		}

		/// <summary> Called after the frame drawing. </summary>
		private void LateUpdate()
		{	if (ParentWidget.Selected) MetricsManager.AddTime(MetricsManager.TaskId.UC1C, Time.deltaTime); }

		#endregion
		
		#region Public Methods
		
		/// <summary> Called upon key pressing. </summary>
		/// <param name="key"> The pressed Key. </param>
		public override void KeyPressed(Key key)
		{
			// Analyze the given key name
			switch (key.name)
			{
			case "Key_OK": ParentWidget.Selected = false; break;
			case "Key_Record":

				// Make sure we are in the correct state
				if (key.CurrentIndex != 0) return;

				// Capture the sound from the microphone
				InputManager.RecordSound("Salute", SoundLength);

				// Change the keys indexes
				KeyRecord.ChangeIndex(1);
				KeyPlay.ChangeIndex(0);

				// Return to the ready state after 3 seconds
				Manager.Invoke(SoundLength, this.SoundReady);

				break;

			case "Key_Play":

				// Make sure we are in the correct state
				if (key.CurrentIndex != 1) return;

				// Play the recorded sound
				OutputManager.PlaySound("Salute", this.transform.position);

				// Change the keys indexes
				KeyRecord.ChangeIndex(2); KeyPlay.ChangeIndex(2);

				// Return to the ready state after 3 seconds
				Manager.Invoke(SoundLength, this.SoundReady);

				break;
			}
		}

		#endregion
	}
}

