using UnityEngine;
using System;
using System.Collections.Generic;
using XUIML.Data;
using XUIML.Model;

namespace XUIML.Unity
{
	/// <summary> The Unity behaviour applied to the Entities associated with a Widget. </summary>
	public class WidgetBehaviour : MonoBehaviour
	{
		#region Public Events
		
		/// <summary> Defines the structure of event handlers </summary>
		public delegate void EventHandler (WidgetBehaviour widget);
		
		/// <summary> Occurs when an node is enabled or disabled. </summary>
		public event EventHandler EnabledStateChanged;

		/// <summary> Occurs when an node is selected or deselected. </summary>
		public event EventHandler SelectedStateChanged;

		#endregion

		#region Public Enumarations

		/// <summary> Enumerates the different activation modes. </summar>
		public enum ActivationMode { None, OnClick, OnPress };

		#endregion

		#region Private Fields

		/// <summary> Indicates if the widget is enabled or not. </summary>
		private bool IsEnabled;

		/// <summary> Indicates if the widget is selected or not. </summary>
		private bool IsSelected;

		/// <summary> The Show animation. </summary>
		private String ShowAnimation;

		/// <summary> The Hide animation. </summary>
		private String HideAnimation;

		/// <summary> The widget that is enabled when the object is selected. </summary>
		private Widget SelectionWidget;

		/// <summary> The selection group to witch this object belongs to. </summary>
		private string SelectionGroup;

		private static Dictionary<string, List<WidgetBehaviour>> SelectionGroups;

		
		#endregion
		
		#region Public Fields

		/// <summary> The GameObject reference that serves as the connector for child widgets. </summary>
		public GameObject Connector;

		/// <summary> The activation mode. </summary>
		public ActivationMode Activation = ActivationMode.OnClick;

		public bool Resizeable = false;

		public string SelectionSound;

		public float MaxTextSize = 0f;

		#endregion

		#region Public Properties
		
		/// <summary> The XUIML Widget node. </summary>
		public Widget WidgetNode { get; private set; }

		public WidgetBehaviour Parent { get; private set; }

		public bool Enabled { get { return IsEnabled; } set { SetEnabled(value); } }

		public bool Selected { get { return IsSelected; } set { SetSelected(value); } }

		/// <summary> The Renderer components of the widget. </summary>
		public List <Renderer> Renderers { get; private set; }

		/// <summary> The Collider components of the widget. </summary>
		public List <Collider> Colliders { get; private set; }

		/// <summary> The light components of the widget. </summary>
		public List <Light> Lights { get; private set; }

		/// <summary> The TextMesh components of the widget. </summary>
		public List <TextMesh> Texts { get; private set; }

		/// <summary> The Image components of the widget. </summary>
		public List <Image> Images { get; private set; }

		/// <summary>The text nodes associated to the widget. </summary>
		public List <Text> TextNodes { get; private set; }

		#endregion

		#region Private Methods

		/// <summary> Gets a list of components associated with the widget. </summary>
		/// <returns> The list of components. </returns>
		/// <param name="go"> The Unity Gameobject to analyze. </param>
		/// <typeparam name="ComponentType"> The type of component to search for.</typeparam>
		private List<ComponentType> ListComponents<ComponentType>(GameObject go) where ComponentType : Component
		{
			// A temporal list of components
			List<ComponentType> list = new List<ComponentType>();

//			if ((go != gameObject) && (go.GetComponent<WidgetBehaviour>())) return list;

			// Check the current gameobject
			ComponentType component = go.GetComponent<ComponentType>();
			if(component != null) list.Add(component);

			// Check the child gameobjects (that only belog to this widget) 
			foreach (Transform child in go.transform) list.AddRange(ListComponents<ComponentType>(child.gameObject));

			// Return the list of components
			return list;
		}

		/// <summary> Sets the enabled state. </summary>
		/// <param name="value">If set to <c>true</c> value.</param>
		private void SetEnabled(bool value, bool forced = false)
		{
			// Make sure the current status is mantained
			if (value == IsEnabled && !forced) return;

			// Make sure the widget is visible
			if (transform.localScale.magnitude == 0) transform.localScale = new Vector3 (0.01f, 0.01f, 0.01f);
			SetVisible(true);

			// Animate and react appropiatelly
			if (value) { Animate("Show"); React("enable"); } else { Animate("Hide"); React("disable"); }

			//if (value) Debug.LogWarning(name + " enabled"); else Debug.LogWarning(name + " disabled");

			// Also update all the child wildgets
			foreach (Transform child in Connector.transform)
			{
				// Do not enable selection widgets
				if (value == true && SelectionWidget != null && child.gameObject == 
				    ((Unity.Element)SelectionWidget.Representation).Existence) continue;

				// Check the child widget behaviour (to avoid activating those with another context)
				WidgetBehaviour childWidgetBehaviour = child.GetComponent<WidgetBehaviour>();
				if (childWidgetBehaviour != null && childWidgetBehaviour.WidgetNode.Context.Conditions.ChildCount == 0)
					childWidgetBehaviour.SetEnabled(value);
			}

			// If the object was selected, select it again
			if(value && Selected) SetSelected(true, true);

			// Update the state value
			IsEnabled = value;

			// Raise the EnabledStateChanged event
			if (EnabledStateChanged != null) EnabledStateChanged(this);
		}

		/// <summary> Shows or hides the widget. </summary>
		/// <param name="value"> The new visible state. </param>
		private void SetVisible(bool value)
		{
			foreach (Renderer r in Renderers) r.enabled = value;
			foreach (Collider c in Colliders) c.enabled = value;
			foreach (Light l in Lights) l.enabled = value;
		}

		/// <summary> Marks the widget has selected or not. </summary>
		/// <param name="value"> The new selected state. </param>
		private void SetSelected(bool value, bool forced = false)
		{
//			Debug.LogWarning("Selected " + name + ":" + value.ToString());

			// Make sure the current status is mantained
			if (value == IsSelected && !forced) return;
			
			if (SelectionWidget != null)
			{
				GameObject selectionObject = ((Unity.Element)SelectionWidget.Representation).Existence;
				WidgetBehaviour selectionBehavior = selectionObject.GetComponent<WidgetBehaviour>();
				if (selectionBehavior != null) { selectionBehavior.SetEnabled(value); }
				Manipulator manipulator = selectionObject.GetComponent<Manipulator>();
				if (manipulator != null && value == true) manipulator.Activate();
			}
			
			// If the widget belongs to a selection group, disable other elements
			if (value == true && SelectionGroup != null)
				foreach (WidgetBehaviour widgetBehaviour in SelectionGroups[SelectionGroup])
					if (widgetBehaviour != this) widgetBehaviour.SetSelected(false);
			
			// Update the state value
			IsSelected = value;

			if (SelectionSound != null) OutputManager.PlaySound(SelectionSound, transform.position, 0.5f);

			// Raise the SelectedStateChanged event
			if (SelectedStateChanged != null) SelectedStateChanged(this);
		}

		/// <summary> Plays an specific animation. </summary>
		/// <param name="animName"> The animation name.</param>
		private void Animate(string animName)
		{ if (animation) foreach (AnimationState s in animation) if(s.name.Contains(animName)) animation.Play(s.name);}

		/// <summary> Called on object creation. </summary>
		private void Awake()
		{
			if (Connector == null) Connector = gameObject;
			this.Renderers = new List<Renderer>();
			this.Colliders = new List<Collider>();
			this.Texts = new List<TextMesh> ();
			this.Images = new List<Image> ();
			this.Lights = new List<Light> ();
			this.TextNodes = new List<Text> ();
			if (SelectionGroups == null) SelectionGroups = new Dictionary<string, List<WidgetBehaviour>>();
		}

		/// <summary> Called upon rendering. </summary>
		private void Update()
		{
			foreach(Renderer childRenderer in Renderers)
				if (childRenderer.enabled && !animation.isPlaying && childRenderer.transform.lossyScale.magnitude == 0)
				{ SetEnabled(false); SetVisible(false); }
		}

		private void OnPressed()
		{ if (Activation == ActivationMode.OnPress){ React("activate"); SetSelected(true); } }

		private void OnClicked()
		{ if (Activation == ActivationMode.OnClick){ React("activate"); SetSelected(true); } }

		private void OnReleased()
		{ if (Activation == ActivationMode.OnPress){ React("activate"); SetSelected(false); } }

		private void OnReleasedWithoutPressing()
		{ if (Activation == ActivationMode.OnPress){ React("activate"); SetSelected(false); } }


		#endregion

		#region Public Methods

		/// <summary> Initializes the widget behaviour. </summary>
		/// <param name="widget"> The Widget node.</param>
		public void Initialize(Widget widget, WidgetBehaviour parent)
		{
			// Store the widget node reference
			this.WidgetNode = widget;
			this.Parent = parent;

			// Get the component list associated to the widget
			this.Renderers = ListComponents <Renderer>(gameObject);
			this.Colliders = ListComponents <Collider>(gameObject);
			this.Texts = ListComponents<TextMesh>(gameObject);
			this.Images = ListComponents<Image>(gameObject);
			this.Lights = ListComponents<Light>(gameObject);

			// Make sure there is a GameObject to connect the child widgets
			if (Connector == null) Connector = gameObject;

			// Start the widget in a disabled, invisible state
			Animate("Hide"); SetVisible(false); 

			// Handle selection properties
			Behavior.Action SelectedAction = widget.Behavior.Actions["initialize"];
			if (SelectedAction != null)
			{
				foreach (Behavior.Action.Reaction reaction in SelectedAction.Reactions)
				{
					if (reaction.Key == "SelectionWidget")
					{
						SelectionWidget = new Widget("SelectionWidget", widget.Widgets);
						SelectionWidget.Type.Value = reaction.Value;
					}
					else if (reaction.Key == "SelectionGroup")
					{
						SelectionGroup = reaction.Value;
						if (!SelectionGroups.ContainsKey(SelectionGroup))
							SelectionGroups.Add(SelectionGroup, new List<WidgetBehaviour>());
						SelectionGroups[SelectionGroup].Add(this);
					}
				}
			}

			SetSelected(false);
 		}

		/// <summary> Reacts the specified action. </summary>
		/// <param name="action"> The action to react to. </param>
		public void React (string action)
		{
			if (WidgetNode == null || WidgetNode.Behavior == null) return;
			
//			Debug.LogWarning ("Reaction: " + this.name + "->" + action);
			
			// Check if the action exists
			if (!WidgetNode.Behavior.Actions.Contains(action)) return;
			
			// Process each associated reaction
			foreach (Behavior.Action.Reaction reaction in WidgetNode.Behavior.Actions[action].Reactions)
			{
				// SPECIAL REACTIONS
				if (reaction.Key == "SaveResults") { Manager.SaveResults(); return; }
				if (reaction.Key == "QuitApplication") {Manager.QuitApplication(); return; }
				if (reaction.Key == "Sound") { OutputManager.PlaySound(reaction.Value,transform.position); continue; }

				Text resource = Manager.GetResource(null, reaction.Key);
				if (resource == null) resource = Manager.GetResource(WidgetNode, reaction.Key);
				if (resource == null) throw new System.Exception("Resource '" + reaction.Key + "' not available for widget " + this.name);

				if (reaction.Operator == "=") resource.Update(reaction.Value);
				else if (reaction.Operator == "+=")
					resource.Update((int.Parse(resource.Value) + int.Parse(reaction.Value)).ToString());
				else if (reaction.Operator == "-=")
					resource.Update((int.Parse(resource.Value) - int.Parse(reaction.Value)).ToString());
			}
		}

		#endregion

	}
}